"""
Run Tests on the Transactions Endpoint
"""

import datetime
import logging
from time import sleep
from typing import List

from lunchable import LunchMoney
from lunchable.models.transactions import (
    TransactionChildObject,
    TransactionInsertObject,
    TransactionObject,
    TransactionSplitObject,
    TransactionUpdateObject,
)
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_get_transactions(lunch_money_obj: LunchMoney):
    """
    Get Transactions and assert they're Transactions
    """
    transactions = lunch_money_obj.get_transactions()
    assert len(transactions) >= 1
    for transaction in transactions:
        assert isinstance(transaction, TransactionObject)
    logger.info("%s Transactions returned", len(transactions))


@lunchable_cassette
def test_get_transaction(lunch_money_obj: LunchMoney):
    """
    Get Transaction (singular) and assert it's a Transaction
    """
    transaction = lunch_money_obj.get_transaction(transaction_id=546434806)
    assert isinstance(transaction, TransactionObject)
    logger.info("Transaction returned: %s", transaction.id)


@lunchable_cassette
def test_insert_transactions(
    lunch_money_obj: LunchMoney, test_transactions: List[TransactionObject]
):
    """
    Insert a Transaction into Lunch Money
    """
    random_note = f"Random Test Description: {datetime.datetime.now()}"
    new_transaction = TransactionInsertObject(
        date=datetime.datetime.now().date(),
        payee="Random Test",
        notes=random_note,
        amount=3.50,
        asset_id=test_transactions[0].asset_id,
    )
    response = lunch_money_obj.insert_transactions(transactions=new_transaction)
    string_ints = [str(integer) for integer in response]
    logger.info("Transactions(s) Created: %s", ", ".join(string_ints))
    for transaction_id in response:
        assert isinstance(transaction_id, int)


@lunchable_cassette
def test_update_transaction(
    lunch_money_obj: LunchMoney, test_transactions: List[TransactionObject]
):
    """
    Update a Transaction in Lunch Money
    """
    transaction_note = f"Updated on {datetime.datetime.now()}"
    transaction_update_obj = TransactionUpdateObject(notes=transaction_note)
    response = lunch_money_obj.update_transaction(
        transaction_id=test_transactions[1].id, transaction=transaction_update_obj
    )
    assert response["updated"] is True


@lunchable_cassette
def test_create_and_delete_transaction_group(
    lunch_money_obj: LunchMoney, test_transactions: List[TransactionObject]
):
    """
    Create a transaction group
    """
    group_id = lunch_money_obj.insert_transaction_group(
        date=datetime.datetime.now().date(),
        payee="Test",
        notes="Test Transaction Group",
        transactions=[test_transactions[1].id, test_transactions[2].id],
    )
    assert isinstance(group_id, int)
    logger.info("Transaction Group created, ID# %s", group_id)
    sleep(1)
    response = lunch_money_obj.remove_transaction_group(transaction_group_id=group_id)
    for transaction_id in response:
        assert isinstance(transaction_id, int)
    logger.info("Transactions part of group: %s", response)


@lunchable_cassette
def test_split_transaction(lunch_money_obj: LunchMoney):
    """
    Try to split a transaction
    """
    transaction_to_split = lunch_money_obj.get_transaction(179018320)
    amount_1 = transaction_to_split.amount / 2
    split_object = TransactionSplitObject(
        date=transaction_to_split.date,
        category_id=transaction_to_split.category_id,
        notes=transaction_to_split.notes,
        amount=amount_1,
    )
    split_object_2 = split_object.model_copy()
    new_split = lunch_money_obj.update_transaction(
        transaction_id=transaction_to_split.id, split=[split_object, split_object_2]
    )
    assert len(new_split["split"]) == 2


@lunchable_cassette
def test_unsplit_transaction(lunch_money_obj: LunchMoney):
    """
    Try to unsplit a transaction
    """
    transaction_ids = [179018299]
    response = lunch_money_obj.unsplit_transactions(
        parent_ids=transaction_ids, remove_parents=True
    )
    assert len(response) == 3


@lunchable_cassette
def test_get_uncleared_transactions(lunch_money_obj: LunchMoney) -> None:
    """
    Get uncleared transactions

    Enum values previously weren't getting JSON encoded
    """
    uncleared_transactions = lunch_money_obj.get_transactions(status="uncleared")
    assert len(uncleared_transactions) >= 1
    for transaction in uncleared_transactions:
        assert isinstance(transaction, TransactionObject)


@lunchable_cassette
def test_204_response(lunch_money_obj: LunchMoney) -> None:
    """
    Test that a 204 response is handled correctly

    This test includes a hand-edited cassette to simulate a 204 response
    """
    test_transaction = TransactionInsertObject(
        date=datetime.date.today(),
        payee="Test",
        notes="Test Transaction Group",
        amount=3.50,
    )
    response = lunch_money_obj.insert_transactions(transactions=test_transaction)
    assert response == []


@lunchable_cassette
def test_get_transaction_group(lunch_money_obj: LunchMoney) -> None:
    """
    Test get_transaction_group
    """
    transaction_group = lunch_money_obj.get_transaction_group(transaction_id=856827078)
    assert isinstance(transaction_group, TransactionObject)
    assert transaction_group.is_group is True
    assert isinstance(transaction_group.children[0], TransactionChildObject)


@lunchable_cassette
def test_get_transactions_no_paginate(lunch_money_obj: LunchMoney) -> None:
    """
    Test get_transactions with no pagination
    """
    transactions = lunch_money_obj.get_transactions(limit=5)
    assert len(transactions) >= 1
    for transaction in transactions:
        assert isinstance(transaction, TransactionObject)



"""
Descriptions for LunchMoney Data Models

This file contains descriptions for the data models used in the LunchMoney API.
Descriptions are seperated into a separate module to keep the code clean and
readable.
"""


class _AssetsDescriptions:
    """
    Descriptions for Assets
    """

    type_name = """
    Primary type of the asset. Must be one of: [employee compensation, cash, vehicle, loan,
    cryptocurrency, investment, other, credit, real estate]
    """
    subtype_name = """
    Optional asset subtype. Examples include: [retirement, checking, savings, prepaid credit card]
    """
    balance = """
    Current balance of the asset in numeric format to 4 decimal places"
    """
    balance_as_of = """
    Date/time the balance was last updated in ISO 8601 extended format
    """
    closed_on = """
    The date this asset was closed (optional)
    """
    currency = """
    Three-letter lowercase currency code of the balance in ISO 4217 format
    """
    created_at = """
    Date/time the asset was created in ISO 8601 extended format
    """
    exclude_transactions = """
    If true, this asset will not show up as an option for assignment when
    creating transactions manually
    """


class _BudgetDescriptions:
    """
    Descriptions for Budget
    """

    category_group_name = "Name of the category group, if applicable"
    is_income = """
    If true, this category is an income category (category properties
    are set in the app via the Categories page)
    """
    exclude_from_budget = """
    If true, this category is excluded from budget (category
    properties are set in the app via the Categories page)
    """
    exclude_from_totals = """
    If true, this category is excluded from totals (category
    properties are set in the app via the Categories page)
    """
    data = """
    For each month with budget or category spending data, there is a data object with the key
    set to the month in format YYYY-MM-DD. For properties, see Data object below.
    """
    config = """
    Object representing the category's budget suggestion configuration
    """


class _CategoriesDescriptions:
    """
    Descriptions for Categories
    """

    name = """
    The name of the category. Must be between 1 and 40 characters.
    """
    description = """
    The description of the category. Must not exceed 140 characters.
    """
    is_income = """
    If true, the transactions in this category will be treated as income.
    """
    exclude_from_budget = """
    If true, the transactions in this category will be excluded from the budget.
    """
    exclude_from_totals = """
    If true, the transactions in this category will be excluded from totals.
    """
    updated_at = """
    The date and time of when the category was last updated (in the ISO
    8601 extended format).
    """
    created_at = """
    The date and time of when the category was created (in the ISO 8601
    extended format).
    """
    is_group = """
    If true, the category is a group that can be a parent to other
    categories.
    """
    group_id = """
    The ID of a category group (or null if the category doesn't belong to
    a category group).
    """
    children = """
    For category groups, this will populate with the categories nested
    within and include id, name, description and created_at fields.
    """
    archived = """
    If true, the category is archived and not displayed in relevant
    areas of the Lunch Money app.
    """
    archived_on = """
    The date and time of when the category was last archived
    (in the ISO 8601 extended format).
    """
    order = """
    Numerical ordering of categories
    """


class _CryptoDescriptions:
    """
    Descriptions for Crypto
    """

    id = """
    Unique identifier for a manual crypto account (no ID for synced accounts)
    """
    zabo_account_id = """
    Unique identifier for a synced crypto account (no ID for manual accounts,
    multiple currencies may have the same zabo_account_id)
    """
    source = """
    `synced` (this account is synced via a wallet, exchange, etc.) or `manual` (this account
    balance is managed manually)
    """
    display_name = "Display name of the crypto asset (as set by user)"
    balance_as_of = """
    Date/time the balance was last updated in ISO 8601 extended format
    """
    status = """
    The current status of the crypto account. Either active or in error.
    """
    created_at = """
    Date/time the asset was created in ISO 8601 extended format
    """


class _PlaidAccountDescriptions:
    """
    Descriptions for Plaid Accounts
    """

    date_linked = """
    Date account was first linked in ISO 8601 extended format
    """
    name = """
    Name of the account. Can be overridden by the user. Field is originally set by Plaid
    """
    type = """
    Primary type of account. Typically one of: [credit, depository, brokerage, cash,
    loan, investment]. This field is set by Plaid and cannot be altered.
    """
    subtype = """
    Optional subtype name of account. This field is set by Plaid and cannot be altered
    """
    mask = """
    Mask (last 3 to 4 digits of account) of account. This field is set by
    Plaid and cannot be altered
    """
    institution_name = """
    Name of institution associated with account. This field is set by
    Plaid and cannot be altered
    """
    status = """
    Denotes the current status of the account within Lunch Money. Must be one of:
    active (Account is active and in good state),
    inactive (Account marked inactive from user. No transactions fetched or
    balance update for this account),
    relink (Account needs to be relinked with Plaid),
    syncing (Account is awaiting first import of transactions),
    error (Account is in error with Plaid),
    not found (Account is in error with Plaid),
    not supported (Account is in error with Plaid)
    """
    last_import = """
    Date of last imported transaction in ISO 8601 extended format (not necessarily
    date of last attempted import)
    """
    balance = """
    Current balance of the account in numeric format to 4 decimal places. This field is
    set by Plaid and cannot be altered
    """
    currency = """
    Currency of account balance in ISO 4217 format. This field is set by Plaid
    and cannot be altered
    """
    balance_last_update = """
    Date balance was last updated in ISO 8601 extended format. This field is set
    by Plaid and cannot be altered
    """
    limit = """
    Optional credit limit of the account. This field is set by Plaid and cannot be altered
    """


class _RecurringExpensesDescriptions:
    """
    Descriptions for Recurring Expenses
    """

    id = """
    Unique identifier for recurring expense
    """
    start_date = """
    Denotes when recurring expense starts occurring in ISO 8601 format.
    If null, then this recurring expense will show up for all time
    before end_date
    """
    end_date = """
    Denotes when recurring expense stops occurring in ISO 8601 format.
    If null, then this recurring expense has no set end date and will
    show up for all months after start_date
    """
    cadence = """
    One of: [monthly, twice a month, once a week, every 3 months, every 4 months,
    twice a year, yearly]
    """
    amount = """
    Amount of the recurring expense in numeric format to 4 decimal places
    """
    currency = """
    Three-letter lowercase currency code for the recurring expense in ISO 4217 format
    """
    description = """
    If any, represents the user-entered description of the recurring expense
    """
    billing_date = """
    Expected billing date for this recurring expense for this month in ISO 8601 format
    """
    type = """
    This can be one of two values: cleared (The recurring expense has been reviewed
    by the user), suggested (The recurring expense is suggested by the system;
    the user has yet to review/clear it)
    """
    original_name = """
    If any, represents the original name of the recurring expense as
    denoted by the transaction that triggered its creation
    """
    source = """
    This can be one of three values: manual (User created this recurring expense
    manually from the Recurring Expenses page), transaction (User created this by
    converting a transaction from the Transactions page), system (Recurring expense
    was created by the system on transaction import). Some older recurring expenses
    may not have a source.
    """
    plaid_account_id = """
    If any, denotes the plaid account associated with the creation of this "
    recurring expense (see Plaid Accounts)
    """
    asset_id = """
    If any, denotes the manually-managed account (i.e. asset) associated with the
    creation of this recurring expense (see Assets)
    """
    transaction_id = """
    If any, denotes the unique identifier for the associated transaction matching
    this recurring expense for the current time period
    """
    category_id = """
    If any, denotes the unique identifier for the associated category to this recurring expense
    """


class _SummarizedTransactionDescriptions:
    """
    Descriptions for Summarized Transaction
    """

    id = """
    Unique identifier for the transaction that matched this recurring item
    """
    date = """
    Date of transaction in ISO 8601 format
    """
    amount = """
    Amount of the transaction in numeric format to 4 decimal places
    """
    currency = """
    Three-letter lowercase currency code of the transaction in ISO 4217 format
    """
    payee = """
    Payee or payer of the recurring item
    """
    category_id = """
    Unique identifier of associated category
    """
    recurring_id = """
    Unique identifier of associated recurring item
    """
    to_base = """
    The amount converted to the user's primary currency. If the multicurrency
    feature is not being used, to_base and amount will be the same.
    """


class _RecurringItemsDescriptions:
    """
    Descriptions for Recurring Items
    """

    id = """
    Unique identifier for recurring item
    """
    start_date = """
    Denotes when recurring item starts occurring in ISO 8601 format.
    If null, then this recurring item will show up for all time before end_date
    """
    end_date = """
    Denotes when recurring item stops occurring in ISO 8601 format.
    If null, then this recurring item has no set end date and will
    show up for all months after start_date
    """
    payee = """
    Payee or payer of the recurring item
    """
    currency = """
    Three-letter lowercase currency code for the recurring item in ISO 4217 format
    """
    created_by = """
    The id of the user who created this recurring item.
    """
    created_at = """
    The date and time of when the recurring item was created (in the ISO 8601
    extended format).
    """
    updated_at = """
    The date and time of when the recurring item was updated (in the ISO 8601 extended format).
    """
    billing_date = """
    Initial date that a transaction associated with this recurring item occured.
    This date is used in conjunction with values of quantity and granularity to
    determine the expected dates of recurring transactions in the period.
    """
    original_name = """
    If any, represents the original name of the recurring item as denoted by
    the transaction that triggered its creation
    """
    description = """
    If any, represents the user-entered description of the recurring item
    """
    plaid_account_id = """
    If any, denotes the plaid account associated with the creation of this
    recurring item (see Plaid Accounts)
    """
    asset_id = """
    If any, denotes the manually-managed account (i.e. asset) associated
    with the creation of this recurring item (see Assets)
    """
    source = """
    This can be one of four values:
    - manual: User created this recurring item manually from the Recurring Items page
    - transaction: User created this by converting a transaction from the Transactions page
    - system: Recurring item was created by the system on transaction import
    - null: Some older recurring items may not have a source.
    """
    notes = """
    If any, the user-entered notes for the recurring item
    """
    amount = """
    Amount of the recurring item in numeric format to 4 decimal places.
    For recurring items with flexible amounts, this is the average of the
    specified min and max amounts.
    """
    category_id = """
    If any, denotes the unique identifier for the associated category to this recurring item
    """
    category_group_id = """
    If any, denotes the unique identifier of associated category group
    """
    is_income = """
    Based on the associated category's property, denotes if the recurring transaction
    is treated as income
    """
    exclude_from_totals = """
    Based on the associated category's property, denotes if the recurring transaction is excluded from totals
    """
    granularity = """
    The unit of time used to define the cadence of the recurring item.
    One of `weeks`, `months`, `years`
    """
    quantity = """
    The number of granular units between each occurrence
    """
    occurrences = """
    An object which contains dates as keys and lists as values. The dates will
    include all the dates in the month that a recurring item is expected, as well
    as the last date in the previous period and the first date in the next period.
    The value for each key is a list of Summarized Transaction Objects that matched
    the recurring item for that date (if any)
    """
    transactions_within_range = """
    A list of all the Summarized Transaction Objects for transactions that that
    have occurred in the query month for the recurring item (if any)
    """
    missing_dates_within_range = """
    A list of date strings when a recurring transaction is expected but has not (yet) occurred.
    """
    date = """
    Denotes the value of the start_date query parameter, or if none was provided, the date when
    the request was made. This indicates the month used by the system when populating the response.
    """
    to_base = """
    The amount converted to the user's primary currency. If the multicurrency feature is not being
    used, to_base and amount will be the same.
    """


class _TransactionInsertDescriptions:
    """
    Descriptions for TransactionInsertObject
    """

    date = """
    Must be in ISO 8601 format (YYYY-MM-DD).
    """
    amount = """
    Numeric value of amount. i.e. $4.25 should be denoted as 4.25.
    """
    category_id = """
    Unique identifier for associated category_id. Category must be associated with
    the same account and must not be a category group.
    """
    currency = """
    Three-letter lowercase currency code in ISO 4217 format. The code sent must exist
    in our database. Defaults to user account's primary currency.
    """
    asset_id = """
    Unique identifier for associated asset (manually-managed account). Asset must be
    associated with the same account.
    """
    recurring_id = """
    Unique identifier for associated recurring expense. Recurring expense must be associated
    with the same account.
    """
    status = """
    Must be either cleared or uncleared. If recurring_id is provided, the status will
    automatically be set to recurring or recurring_suggested depending on the type of
    recurring_id. Defaults to uncleared.
    """
    external_id = """
    User-defined external ID for transaction. Max 75 characters. External IDs must be
    unique within the same asset_id.
    """
    tags = """
    Passing in a number will attempt to match by ID. If no matching tag ID is found, an error
    will be thrown. Passing in a string will attempt to match by string. If no matching tag
    name is found, a new tag will be created.
    """


class _TransactionUpdateDescriptions:
    """
    Descriptions for TransactionUpdateObject
    """

    date = """
    Must be in ISO 8601 format (YYYY-MM-DD).
    """
    category_id = """
    Unique identifier for associated category_id. Category must be associated
    with the same account and must not be a category group.
    """
    amount = """
    You may only update this if this transaction was not created from an automatic
    import, i.e. if this transaction is not associated with a plaid_account_id
    """
    currency = """
    You may only update this if this transaction was not created from an automatic
    import, i.e. if this transaction is not associated with a plaid_account_id.
    Defaults to user account's primary currency.
    """
    asset_id = """
    Unique identifier for associated asset (manually-managed account). Asset must be
    associated with the same account. You may only update this if this transaction was
    not created from an automatic import, i.e. if this transaction is not associated
    with a plaid_account_id
    """
    recurring_id = """
    Unique identifier for associated recurring expense. Recurring expense must
    be associated with the same account.
    """
    status = """
    Must be either cleared or uncleared. Defaults to uncleared If recurring_id is
    provided, the status will automatically be set to recurring or recurring_suggested
    depending on the type of recurring_id. Defaults to uncleared.
    """
    external_id = """
    User-defined external ID for transaction. Max 75 characters. External IDs must be
    unique within the same asset_id. You may only update this if this transaction was
    not created from an automatic import, i.e. if this transaction is not associated
    with a plaid_account_id
    """
    tags = """
    Passing in a number will attempt to match by ID. If no matching tag ID is found,
    an error will be thrown. Passing in a string will attempt to match by string.
    If no matching tag name is found, a new tag will be created.
    """


class _TransactionSplitDescriptions:
    """
    Descriptions for TransactionSplitObject
    """

    date = "Must be in ISO 8601 format (YYYY-MM-DD)."
    category_id = """
    Unique identifier for associated category_id. Category must be associated
    with the same account.
    """
    notes = "Transaction Split Notes."
    amount = """
    Individual amount of split. Currency will inherit from parent transaction. All
    amounts must sum up to parent transaction amount.
    """


class _TransactionDescriptions:
    """
    Descriptions for TransactionObject
    """

    id = """
    Unique identifier for transaction
    """
    date = """
    Date of transaction in ISO 8601 format
    """
    payee = """
    Name of payee. If recurring_id is not null, this field will show the payee
    of associated recurring expense instead of the original transaction payee
    """
    amount = """
    Amount of the transaction in numeric format to 4 decimal places
    """
    currency = """
    Three-letter lowercase currency code of the transaction in ISO 4217 format
    """
    to_base = """
    The amount converted to the user's primary currency. If the multicurrency
    feature is not being used, to_base and amount will be the same.
    """
    category_id = """
    Unique identifier of associated category
    """
    category_name = """
    Name of category associated with transaction
    """
    category_group_id = """
    Unique identifier of associated category group, if any
    """
    category_group_name = """
    Name of category group associated with transaction, if any
    """
    is_income = """
    Based on the associated category's property, denotes if transaction is
    treated as income
    """
    exclude_from_budget = """
    Based on the associated category's property, denotes if transaction is
    excluded from budget
    """
    exclude_from_totals = """
    Based on the associated category's property, denotes if transaction is
    excluded from totals
    """
    created_at = """
    The date and time of when the transaction was created (in the ISO 8601
    extended format).
    """
    updated_at = """
    The date and time of when the transaction was last updated (in the ISO 8601
    extended format).
    """
    status = """
    One of the following: <ul>
        <li>cleared: User has reviewed the transaction</li>
        <li>uncleared: User has not yet reviewed the transaction</li>
        <li>recurring: Transaction is linked to a recurring expense</li>
        <li>recurring_suggested: Transaction is listed as a suggested transaction
            for an existing recurring expense.</li>
        <li>pending: Imported transaction is marked as pending. This should be a
            temporary state.</li>
    </ul>
    User intervention is required to change this to recurring.
    """
    is_pending = """
    Denotes if transaction is pending (not posted)
    """
    notes = """
    User-entered transaction notes If recurring_id is not null, this field will
    be description of associated recurring expense
    """
    original_name = """
    The transactions original name before any payee name updates. For synced
    transactions, this is the raw original payee name from your bank.
    """
    recurring_id = """
    Unique identifier of associated recurring item
    """
    recurring_payee = """
    Payee name of associated recurring item
    """
    recurring_description = """
    Description of associated recurring item
    """
    recurring_cadence = """
    Cadence of associated recurring item (one of `once a week`, `every 2 weeks`,
    `twice a month`, `monthly`, `every 2 months`, `every 3 months`, `every 4 months`,
    `twice a year`, `yearly`)
    """
    recurring_type = """
    Type of associated recurring (one of `cleared`, `suggested`, `dismissed`)
    """
    recurring_amount = """
    Amount of associated recurring item
    """
    recurring_currency = """
    Currency of associated recurring item
    """
    parent_id = """
    Exists if this is a split transaction. Denotes the transaction ID of the
    original transaction. Note that the parent transaction is not returned in
    this call.
    """
    has_children = """
    True if this transaction is a parent transaction and is split into 2 or
    more other transactions
    """
    group_id = """
    Exists if this transaction is part of a group. Denotes the parent’s
    transaction ID
    """
    is_group = """
    True if this transaction represents a group of transactions. If so, amount
    and currency represent the totalled amount of transactions bearing this
    transaction’s id as their group_id. Amount is calculated based on the
    user’s primary currency.
    """
    asset_id = """
    Unique identifier of associated manually-managed account (see Assets)
    Note: plaid_account_id and asset_id cannot both exist for a transaction
    """
    asset_institution_name = """
    Institution name of associated manually-managed account
    """
    asset_name = """
    Name of associated manually-managed account
    """
    asset_display_name = """
    Display name of associated manually-managed account
    """
    asset_status = """
    Status of associated manually-managed account (one of `active`, `closed`)
    """
    plaid_account_id = """
    Unique identifier of associated Plaid account (see Plaid Accounts) Note:
    plaid_account_id and asset_id cannot both exist for a transaction
    """
    plaid_account_name = """
    Name of associated Plaid account
    """
    plaid_account_mask = """
    Mask of associated Plaid account
    """
    institution_name = """
    Institution name of associated Plaid account
    """
    plaid_account_display_name = """
    Display name of associated Plaid account
    """
    plaid_metadata = """
    Metadata associated with imported transaction from Plaid
    """
    source = """
    Source of the transaction (one of `api`, `csv`, `manual`,`merge`,`plaid`,
    `recurring`,`rule`,`user`)
    """
    display_name = """
    Display name for payee for transaction based on whether or not it is
    linked to a recurring item. If linked, returns `recurring_payee` field.
    Otherwise, returns the `payee` field.
    """
    display_notes = """
    Display notes for transaction based on whether or not it is linked to a
    recurring item. If linked, returns `recurring_notes` field. Otherwise,
    returns the `notes` field.
    """
    account_display_name = """
    Display name for associated account (manual or Plaid). If this is a synced
    account, returns `plaid_account_display_name` or `asset_display_name`.
    """
    tags = """
    Array of Tag objects
    """
    external_id = """
    User-defined external ID for any manually-entered or imported transaction.
    External ID cannot be accessed or changed for Plaid-imported transactions.
    External ID must be unique by asset_id. Max 75 characters.
    """
    children = """
    Array of Transaction objects. Only exists if this transaction is a parent
    transaction and is split into 2 or more other transactions. Child transactions
    do not contain all of the same fields as parent transactions.
    """
    formatted_date = """
    Date of transaction in user's preferred format
    """



"""
Run Tests on the Assets Endpoint
"""

import datetime
import logging

import pytest

from lunchable import LunchMoney
from lunchable.models.assets import AssetsObject
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@pytest.fixture
def lunchmoney_asset() -> AssetsObject:
    """
    Static lunchable.models.AssetsObject

    Returns
    -------
    AssetsObject
    """
    splitwise_asset = AssetsObject(
        id=78214,
        type_name="cash",
        subtype_name="digital wallet (paypal, venmo)",
        name="Splitwise Balance",
        display_name="Splitwise",
        balance=-1.0,
        balance_as_of=datetime.datetime(
            2021, 8, 28, 16, 6, 35, tzinfo=datetime.timezone.utc
        ),
        closed_on=None,
        currency="usd",
        institution_name="Splitwise",
        created_at=datetime.datetime(
            2021, 8, 28, 16, 6, 2, 701000, tzinfo=datetime.timezone.utc
        ),
    )
    return splitwise_asset


@lunchable_cassette
def test_get_assets(lunch_money_obj: LunchMoney):
    """
    Get Assets and Assert that they're assets
    """
    assets = lunch_money_obj.get_assets()
    assert len(assets) >= 1
    for asset in assets:
        assert isinstance(asset, AssetsObject)
    logger.info("%s Assets returned", len(assets))


@lunchable_cassette
def test_update_asset(lunch_money_obj: LunchMoney, lunchmoney_asset: AssetsObject):
    """
    Update an Asset
    """
    response = lunch_money_obj.update_asset(asset_id=lunchmoney_asset.id, balance=5.20)
    assert isinstance(response, AssetsObject)
    logger.info(response)


@lunchable_cassette
def test_create_asset(lunch_money_obj: LunchMoney):
    """
    Create an Asset
    """
    response = lunch_money_obj.insert_asset(
        type_name="cash",
        name="test-account-1",
        subtype_name=None,
        display_name="Test Account #1",
        balance=4.20,
        currency="usd",
        institution_name="Test Institution",
    )
    assert isinstance(response, AssetsObject)
    logger.info(response)



"""
Dynamic Logging Configuration
"""

import logging
from os import getenv
from typing import Optional, Tuple, Union

import click
from rich.console import Console
from rich.logging import RichHandler

LOG_HANDLER = getenv("LOG_HANDLER", "rich").lower()


def get_log_handler(
    log_level: Optional[int] = None,
) -> Tuple[logging.Handler, Union[int, str]]:
    """
    Determine which logging handler should be used

    Parameters
    ----------
    log_level: Optional[int]
        Which logging level should be used. If none is provided the LOG_LEVEL environment
        variable will be used, defaulting to "INFO".

    Returns
    -------
    Tuple[logging.Handler, Union[int, str]]
    """
    if log_level is None:
        log_level = logging.getLevelName(getenv("LOG_LEVEL", "INFO").upper())
    rich_handler = RichHandler(
        level=log_level,
        rich_tracebacks=True,
        omit_repeated_times=False,
        show_path=False,
        tracebacks_suppress=[click],
        console=Console(stderr=True),
    )
    httpx_logger = logging.getLogger("httpx")
    if log_level != logging.DEBUG:
        httpx_logger.setLevel(logging.WARNING)
    python_handler = logging.StreamHandler()
    python_formatter = logging.Formatter("%(asctime)s [%(levelname)8s]: %(message)s")
    python_handler.setFormatter(python_formatter)
    python_handler.setLevel(log_level)
    _log_dict = {
        "rich": rich_handler,
        "python": python_handler,
    }
    if getenv("PYTEST_CURRENT_TEST", None) is not None:
        handler = "python"
    else:
        handler = LOG_HANDLER
    log_handler: logging.Handler = _log_dict.get(handler, rich_handler)
    return log_handler, log_level


def set_up_logging(log_level: Optional[int] = None) -> None:
    """
    Set Up a Root Logger

    Parameters
    ----------
    log_level: Optional[int]
        Which logging level should be used. If none is provided the LOG_LEVEL environment
        variable will be used, defaulting to "INFO".
    """
    log_handler, level_to_log = get_log_handler(log_level=log_level)
    logging.root.handlers = [log_handler]
    if isinstance(log_handler, RichHandler):
        rich_formatter = logging.Formatter(
            datefmt="[%Y-%m-%d %H:%M:%S]", fmt="%(message)s"
        )
        logging.root.handlers[0].setFormatter(rich_formatter)
        level_to_log = logging.NOTSET
    logging.root.setLevel(level_to_log)



"""
Base Pydantic Object for Containers
"""

from pydantic import BaseModel, ConfigDict


class LunchableModel(BaseModel):
    """
    Hashable Pydantic Model
    """

    model_config = ConfigDict(extra="allow")

    def __hash__(self) -> int:
        """
        Hash Method for Pydantic BaseModels
        """
        return hash((type(self), *tuple(self.__dict__.values())))



"""
Run Tests on the Crypto Endpoint
"""

import logging

from lunchable import LunchMoney
from lunchable.models.crypto import CryptoObject
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_get_crypto(lunch_money_obj: LunchMoney):
    """
    Get Crypto and assert its Crypto
    """
    cryptos = lunch_money_obj.get_crypto()
    assert len(cryptos) >= 1
    for crypto in cryptos:
        assert isinstance(crypto, CryptoObject)
    logger.info("%s Crypto Accounts returned", len(cryptos))


@lunchable_cassette
def test_update_crypto(lunch_money_obj: LunchMoney):
    """
    Update a Crypto Object
    """
    crypto = lunch_money_obj.update_crypto(crypto_id=7286, balance=0.50)
    assert isinstance(crypto, CryptoObject)
    logger.info("Crypto Asset Updated: %s", crypto.id)



"""
Base Classes for lunchable-app
"""

from __future__ import annotations

import datetime
import functools
import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, ClassVar, Dict, List, Tuple, Type, TypeVar, overload

from pydantic import BaseModel, Field

from lunchable import LunchMoney
from lunchable.models import (
    AssetsObject,
    CategoriesObject,
    CryptoObject,
    LunchableModel,
    PlaidAccountObject,
    TagsObject,
    TransactionObject,
    UserObject,
)

logger = logging.getLogger(__name__)

LunchableModelType = TypeVar("LunchableModelType", bound=LunchableModel)


class LunchableData(BaseModel):
    """
    Data Container for Lunchable App Data
    """

    plaid_accounts: Dict[int, PlaidAccountObject] = Field(
        default_factory=dict, description="Plaid Accounts"
    )
    transactions: Dict[int, TransactionObject] = Field(
        default_factory=dict, description="Transactions"
    )
    categories: Dict[int, CategoriesObject] = Field(
        default_factory=dict, description="Categories"
    )
    assets: Dict[int, AssetsObject] = Field(default_factory=dict, description="Assets")
    tags: Dict[int, TagsObject] = Field(default_factory=dict, description="Tags")
    crypto: Dict[int, CryptoObject] = Field(default_factory=dict, description="Crypto")
    user: UserObject = Field(
        UserObject(
            user_id=0, user_name="", user_email="", account_id=0, budget_name=""
        ),
        description="User",
    )

    @property
    def asset_map(self) -> Dict[int, PlaidAccountObject | AssetsObject]:
        """
        Asset Mapping Across Plaid Accounts and Assets

        Returns
        -------
        Dict[int, Union[PlaidAccountObject, AssetsObject]]
        """
        return {
            **self.plaid_accounts,
            **self.assets,
        }

    @property
    def plaid_accounts_list(self) -> List[PlaidAccountObject]:
        """
        List of Plaid Accounts

        Returns
        -------
        List[PlaidAccountObject]
        """
        return list(self.plaid_accounts.values())

    @property
    def assets_list(self) -> List[AssetsObject]:
        """
        List of Assets

        Returns
        -------
        List[AssetsObject]
        """
        return list(self.assets.values())

    @property
    def transactions_list(self) -> List[TransactionObject]:
        """
        List of Transactions

        Returns
        -------
        List[TransactionObject]
        """
        return list(self.transactions.values())

    @property
    def categories_list(self) -> List[CategoriesObject]:
        """
        List of Categories

        Returns
        -------
        List[CategoriesObject]
        """
        return list(self.categories.values())

    @property
    def tags_list(self) -> List[TagsObject]:
        """
        List of Tags

        Returns
        -------
        List[TagsObject]
        """
        return list(self.tags.values())

    @property
    def crypto_list(self) -> List[CryptoObject]:
        """
        List of Crypto

        Returns
        -------
        List[CryptoObject]
        """
        return list(self.crypto.values())


class BaseLunchableApp(ABC):
    """
    Abstract Base Class for Lunchable Apps
    """

    @property
    def _lunchable_data_mapping(
        self,
    ) -> Dict[
        Type[LunchableModel],
        Tuple[str, Callable[[], List[LunchableModel] | LunchableModel]],
    ]:
        """
        Mapping of Lunchable Objects to their Data Collecting Info
        """
        return {
            PlaidAccountObject: ("plaid_accounts", self.lunch.get_plaid_accounts),
            TransactionObject: ("transactions", self.lunch.get_transactions),
            CategoriesObject: ("categories", self.lunch.get_categories),
            AssetsObject: ("assets", self.lunch.get_assets),
            TagsObject: ("tags", self.lunch.get_tags),
            UserObject: ("user", self.lunch.get_user),
            CryptoObject: ("crypto", self.lunch.get_crypto),
        }

    def __init__(self, access_token: str | None = None):
        """
        Lunchable App Initialization

        Parameters
        ----------
        access_token: Optional[str]
            Lunchmoney Developer API Access Token. Inherited from
            `LUNCHMONEY_ACCESS_TOKEN` environment variable if not provided
        """
        self.lunch = LunchMoney(access_token=access_token)
        self.data = LunchableData()

    @property
    @abstractmethod
    def lunchable_models(self) -> List[Type[LunchableModel]]:
        """
        Every LunchableApp should define which data objects it depends on

        Returns
        -------
        List[LunchableDataModel]
        """

    @overload
    def refresh(self, model: Type[UserObject], **kwargs: Any) -> UserObject:
        ...

    @overload
    def refresh(
        self, model: Type[LunchableModelType], **kwargs: Any
    ) -> Dict[int, LunchableModelType]:
        ...

    def refresh(
        self, model: Type[LunchableModel], **kwargs: Any
    ) -> LunchableModel | Dict[int, LunchableModel]:
        """
        Refresh a Lunchable Model

        Parameters
        ----------
        model: Type[LunchableModel]
            Type of Lunchable Model to refresh
        kwargs: Any
            Additional keyword arguments to pass to the function that
            fetches the data.

        Returns
        -------
        LunchableModel | Dict[int, LunchableModel]
            Unless you're requesting the `UserObject`, this method will return a
            dictionary of the refreshed data, keyed by the object's ID.

        Examples
        --------
        ```python
        from typing import Dict

        from lunchable.models import CategoriesObject
        from lunchable.plugins import LunchableApp

        app = LunchableApp()
        categories: Dict[int, CategoriesObject] = app.refresh(CategoriesObject)
        ```
        """
        try:
            attr_name, data_getter = self._lunchable_data_mapping[model]
            fetch_data_function = functools.partial(data_getter, **kwargs)
        except KeyError as e:
            msg = f"Model not supported by Lunchable App: {model.__name__}"
            raise NotImplementedError(msg) from e
        fetched_data = fetch_data_function()
        if isinstance(fetched_data, UserObject):
            data_mapping = fetched_data
        else:
            data_mapping = {item.id: item for item in fetched_data}  # type: ignore[assignment]
        setattr(self.data, attr_name, data_mapping)
        return data_mapping

    def refresh_data(self, models: List[Type[LunchableModel]] | None = None) -> None:
        """
        Refresh the data in the Lunchable App

        Parameters
        ----------
        models: List[Type[LunchableModel]] | None
            Explicit list of Lunchable Models to refresh. If not provided,
            all models defined in will be refreshed (which by default is
            all of them except for transactions)

        Examples
        --------
        ```python
        from typing import Dict

        from lunchable.models import PlaidAccountObject
        from lunchable.plugins import LunchableApp

        app = LunchableApp()
        app.refresh_data()
        plaid_accounts: Dict[int, PlaidAccountObject] = app.data.plaid_accounts
        ```

        ```python
        from typing import Dict

        from lunchable.models import AssetsObject
        from lunchable.plugins import LunchableApp

        app = LunchableApp()
        app.refresh_data(models=[AssetsObject])
        assets: Dict[int, AssetsObject] = app.data.assets
        ```
        """
        refresh_models = models or self.lunchable_models
        for model in refresh_models:
            self.refresh(model)

    def refresh_transactions(
        self,
        start_date: datetime.date | datetime.datetime | str | None = None,
        end_date: datetime.date | datetime.datetime | str | None = None,
        tag_id: int | None = None,
        recurring_id: int | None = None,
        plaid_account_id: int | None = None,
        category_id: int | None = None,
        asset_id: int | None = None,
        group_id: int | None = None,
        is_group: bool | None = None,
        status: str | None = None,
        debit_as_negative: bool | None = None,
        pending: bool | None = None,
        params: Dict[str, Any] | None = None,
    ) -> Dict[int, TransactionObject]:
        """
        Refresh App data with the latest transactions

        Parameters
        ----------
        start_date: Optional[Union[datetime.date, datetime.datetime, str]]
            Denotes the beginning of the time period to fetch transactions for.
            Defaults to beginning of current month. Required if end_date exists.
            Format: YYYY-MM-DD.
        end_date: Optional[Union[datetime.date, datetime.datetime, str]]
            Denotes the end of the time period you'd like to get transactions for.
            Defaults to end of current month. Required if start_date exists.
        tag_id: Optional[int]
            Filter by tag. Only accepts IDs, not names.
        recurring_id: Optional[int]
            Filter by recurring expense
        plaid_account_id: Optional[int]
            Filter by Plaid account
        category_id: Optional[int]
            Filter by category. Will also match category groups.
        asset_id: Optional[int]
            Filter by asset
        group_id: Optional[int]
            Filter by group_id (if the transaction is part of a specific group)
        is_group: Optional[bool]
            Filter by group (returns transaction groups)
        status: Optional[str]
            Filter by status (Can be cleared or uncleared. For recurring
            transactions, use recurring)
        debit_as_negative: Optional[bool]
            Pass in true if you'd like expenses to be returned as negative amounts and
            credits as positive amounts. Defaults to false.
        pending: Optional[bool]
            Pass in true if you'd like to include imported transactions with a
            pending status.
        params: Optional[dict]
            Additional Query String Params

        Returns
        -------
        Dict[int, TransactionObject]

        Examples
        --------
        ```python
        from typing import Dict

        from lunchable.models import TransactionObject
        from lunchable.plugins import LunchableApp

        app = LunchableApp()
        transactions: Dict[int, TransactionObject] = app.refresh_transactions(
            start_date="2021-01-01", end_date="2021-01-31"
        )
        ```
        """
        transactions = self.lunch.get_transactions(
            start_date=start_date,
            end_date=end_date,
            tag_id=tag_id,
            recurring_id=recurring_id,
            plaid_account_id=plaid_account_id,
            category_id=category_id,
            asset_id=asset_id,
            group_id=group_id,
            is_group=is_group,
            status=status,
            debit_as_negative=debit_as_negative,
            pending=pending,
            params=params,
        )
        transaction_map = {item.id: item for item in transactions}
        self.data.transactions.update(transaction_map)
        return transaction_map

    def clear_transactions(self) -> None:
        """
        Clear Transactions from the App
        """
        self.data.transactions.clear()


class LunchableApp(BaseLunchableApp):
    """
    Pre-Built Lunchable App

    This app comes with a `data` property which represents all the base data
    the app should need.

    This app comes with a `refresh_data` method which will refresh all of the
    data in the app, except for transactions. To refresh transactions, use the
    `refresh_transactions` method.
    """

    lunchable_models: ClassVar[List[Type[LunchableModel]]] = [
        PlaidAccountObject,
        CategoriesObject,
        AssetsObject,
        TagsObject,
        UserObject,
        CryptoObject,
    ]



"""
Pytest Fixtures Shared Across all Unit Tests
"""

import datetime
import os
import pathlib
from typing import List

import pytest
from vcr import VCR

from lunchable import LunchMoney
from lunchable.models import TransactionObject

obscure_start_date_object = datetime.datetime(year=2022, month=11, day=1)
beginning_of_this_month = datetime.datetime.now().replace(day=1)
module_scope = pytest.fixture(scope="module")


@pytest.fixture(autouse=True)
def set_test_env_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Set Environment Variables for Testing if they are not already set
    """
    testing_env_vars = [
        "LUNCHMONEY_ACCESS_TOKEN",
        "PUSHOVER_USER_KEY",
        "SPLITWISE_API_KEY",
        "SPLITWISE_CONSUMER_KEY",
        "SPLITWISE_CONSUMER_SECRET",
    ]
    for env_var in testing_env_vars:
        if not os.getenv(env_var):
            monkeypatch.setenv(env_var, f"{env_var}_PLACEHOLDER")


@pytest.fixture
def obscure_start_date() -> datetime.datetime:
    """
    An Obscure Hardcoded Date

    Returns
    -------
    datetime.datetime
    """
    return obscure_start_date_object


@pytest.fixture
def lunch_money_obj() -> LunchMoney:
    """
    Static LunchMoney Instance

    Returns
    -------
    LunchMoney
    """
    lunch_money_obj = LunchMoney()
    assert isinstance(lunch_money_obj, LunchMoney)
    return lunch_money_obj


@pytest.fixture
def test_transactions() -> List[TransactionObject]:
    """
    Load some example transactions
    """
    transaction_dict_1 = {
        "amount": 1.0,
        "asset_id": 49335,
        "category_id": 658761,
        "currency": "usd",
        "date": "2021-09-19",
        "external_id": None,
        "fees": None,
        "group_id": None,
        "id": 546434801,
        "is_group": False,
        "notes": "Test Transaction 1",
        "original_name": "Test 1",
        "parent_id": None,
        "payee": "Test 1",
        "plaid_account_id": None,
        "price": None,
        "quantity": None,
        "status": "uncleared",
        "subtype": None,
        "tags": None,
        "type": None,
        "created_at": "2021-09-19T20:00:00.000Z",
        "updated_at": "2021-09-19T20:00:00.000Z",
    }
    transaction_dict_2 = {
        "amount": 2.0,
        "asset_id": 23043,
        "category_id": 229146,
        "currency": "usd",
        "date": "2021-09-19",
        "external_id": None,
        "fees": None,
        "group_id": None,
        "id": 546452296,
        "is_group": False,
        "notes": "Test Transaction 2",
        "original_name": "Test 2",
        "parent_id": None,
        "payee": "Test 2",
        "plaid_account_id": None,
        "price": None,
        "quantity": None,
        "status": "uncleared",
        "subtype": None,
        "tags": None,
        "type": None,
        "created_at": "2021-09-19T20:00:00.000Z",
        "updated_at": "2021-09-19T20:00:00.000Z",
    }
    transaction_dict_3 = {
        "amount": 3.0,
        "asset_id": 23043,
        "category_id": 229140,
        "currency": "usd",
        "date": "2021-09-19",
        "external_id": None,
        "fees": None,
        "group_id": None,
        "id": 546434806,
        "is_group": False,
        "notes": "Test Transaction 3",
        "original_name": "Test 3",
        "parent_id": None,
        "payee": "Test 3",
        "plaid_account_id": None,
        "price": None,
        "quantity": None,
        "status": "uncleared",
        "subtype": None,
        "tags": None,
        "type": None,
        "created_at": "2021-09-19T20:00:00.000Z",
        "updated_at": "2021-09-19T20:00:00.000Z",
    }
    transaction_1 = TransactionObject.model_validate(transaction_dict_1)
    transaction_2 = TransactionObject.model_validate(transaction_dict_2)
    transaction_3 = TransactionObject.model_validate(transaction_dict_3)
    return [transaction_1, transaction_2, transaction_3]


###########################################################
# VCR Configuration : Offload Epic API Calls to Cassettes #
###########################################################


def path_transformer(path: str) -> str:
    """
    Cassette Path Transformer
    """
    suffix = ".yaml"
    if not path.endswith(suffix):
        path = path + suffix
    cassette_path = pathlib.Path(path)
    cassette_path = cassette_path.parent / "cassettes" / cassette_path.name
    return str(cassette_path)


vcr = VCR(
    filter_headers=(("authorization", "XXXXXXXXXX"),),
    filter_query_parameters=(("user", "XXXXXXXXXX"), ("token", "XXXXXXXXXX")),
    path_transformer=path_transformer,
    record_mode=os.getenv("VCR_RECORD_MODE", "once"),
)

# Decorator Object to Use pyvcr Cassettes on Unit Tests
# pass `--vcr-record=none` to pytest CI runs to ensure new cassettes are generated
lunchable_cassette = vcr.use_cassette



"""
Run Tests on the Miscellaneous Endpoints
"""

import logging

from lunchable import LunchMoney
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_get_user(lunch_money_obj: LunchMoney):
    """
    Get Me Object
    """
    me = lunch_money_obj.get_user()
    assert isinstance(me.user_id, int)
    assert isinstance(me.user_email, str)
    assert isinstance(me.budget_name, str)
    assert me.api_key_label == "Lunchable"



"""
Lunch Money - Categories

https://lunchmoney.dev/#categories
"""

from __future__ import annotations

import datetime
import json
import logging
from enum import Enum
from typing import List, Optional, Union

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.exceptions import LunchMoneyError
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient
from lunchable.models._descriptions import _CategoriesDescriptions

logger = logging.getLogger(__name__)


class ModelCreateCategory(LunchableModel):
    """
    https://lunchmoney.dev/#create-category
    """

    name: str
    description: Optional[str] = None
    is_income: Optional[bool] = None
    exclude_from_budget: Optional[bool] = None
    exclude_from_totals: Optional[bool] = None
    archived: Optional[bool] = None
    group_id: Optional[int] = None


class CategoryChild(LunchableModel):
    """
    Child Entry on the Category Object
    """

    id: int
    name: str = Field(min_length=1, max_length=40)
    description: Optional[str] = Field(None, max_length=140)
    created_at: Optional[datetime.datetime] = None


class CategoriesObject(LunchableModel):
    """
    Lunch Money Spending Categories

    https://lunchmoney.dev/#categories-object
    """

    id: int = Field(description="A unique identifier for the category.")
    name: str = Field(
        min_length=1,
        max_length=40,
        description=_CategoriesDescriptions.name,
    )
    description: Optional[str] = Field(
        None, max_length=140, description=_CategoriesDescriptions.description
    )
    is_income: bool = Field(description=_CategoriesDescriptions.is_income)
    exclude_from_budget: bool = Field(
        description=_CategoriesDescriptions.exclude_from_budget
    )
    exclude_from_totals: bool = Field(
        description=_CategoriesDescriptions.exclude_from_totals
    )
    archived: bool = Field(False, description=_CategoriesDescriptions.archived)
    archived_on: Optional[datetime.datetime] = Field(
        None, description=_CategoriesDescriptions.archived_on
    )
    updated_at: Optional[datetime.datetime] = Field(
        None, description=_CategoriesDescriptions.updated_at
    )
    created_at: Optional[datetime.datetime] = Field(
        None, description=_CategoriesDescriptions.created_at
    )
    is_group: bool = Field(description=_CategoriesDescriptions.is_group)
    group_id: Optional[int] = Field(None, description=_CategoriesDescriptions.group_id)
    order: Optional[int] = Field(None, description=_CategoriesDescriptions.order)
    children: Optional[List[Union[CategoriesObject, CategoryChild]]] = Field(
        None,
        description=_CategoriesDescriptions.children,
    )


class _CategoriesParamsPut(LunchableModel):
    """
    https://lunchmoney.dev/#update-category
    """

    name: Optional[str] = Field(None, min_length=1, max_length=40)
    description: Optional[str] = Field(None, max_length=140)
    is_income: Optional[bool] = None
    exclude_from_budget: Optional[bool] = None
    exclude_from_totals: Optional[bool] = None
    archived: Optional[bool] = None
    group_id: Optional[int] = None


class _CategoriesParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#create-category-group
    """

    name: str
    description: Optional[str] = None
    is_income: Optional[bool] = None
    exclude_from_budget: Optional[bool] = None
    exclude_from_totals: Optional[bool] = None
    category_ids: Optional[List[int]] = None
    new_categories: Optional[List[str]] = None


class _CategoriesAddParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#add-to-category-group
    """

    category_ids: Optional[List[int]] = None
    new_categories: Optional[List[str]] = None


class CategoriesFormatEnum(str, Enum):
    """
    Format Enum
    """

    nested: str = "nested"
    flattened: str = "flattened"


class _GetCategoriesParams(LunchableModel):
    """
    https://lunchmoney.dev/#get-all-categories
    """

    format: Optional[CategoriesFormatEnum] = None


class CategoriesClient(LunchMoneyAPIClient):
    """
    Lunch Money Categories Interactions
    """

    def get_categories(
        self, format: str | CategoriesFormatEnum | None = None
    ) -> List[CategoriesObject]:
        """
        Get Spending categories

        Use this endpoint to get a list of all categories associated with the user's account.
        https://lunchmoney.dev/#get-all-categories

        Parameters
        ----------
        format: str | CategoriesFormatEnum | None
            Can either `flattened` or `nested`. If `flattened`, returns a singular
            array of categories, ordered alphabetically. If `nested`, returns
            top-level categories (either category groups or categories not part
            of a category group) in an array. Subcategories are nested within
            the category group under the property `children`. Defaults to None
            which will return a `flattened` list of categories.

        Returns
        -------
        List[CategoriesObject]
        """
        response_data = self.make_request(
            method=self.Methods.GET,
            url_path=APIConfig.LUNCHMONEY_CATEGORIES,
            params=_GetCategoriesParams(format=format).model_dump(exclude_none=True),
        )
        categories = response_data["categories"]
        category_objects = [
            CategoriesObject.model_validate(item) for item in categories
        ]
        return category_objects

    def insert_category(
        self,
        name: str,
        description: Optional[str] = None,
        is_income: Optional[bool] = None,
        exclude_from_budget: Optional[bool] = None,
        exclude_from_totals: Optional[bool] = None,
        archived: Optional[bool] = None,
        group_id: Optional[int] = None,
    ) -> int:
        """
        Create a Spending Category

        Use this to create a single category
        https://lunchmoney.dev/#create-category

        Parameters
        ----------
        name: str
            Name of category. Must be between 1 and 40 characters.
        description: Optional[str]
            Description of category. Must be less than 140 categories. Defaults to None.
        is_income: Optional[bool]
            Whether or not transactions in this category should be treated as income.
            Defaults to False.
        exclude_from_budget: Optional[bool]
            Whether or not transactions in this category should be excluded from budgets.
            Defaults to False.
        exclude_from_totals: Optional[bool]
            Whether or not transactions in this category should be excluded from
            calculated totals. Defaults to False.
        archived: Optional[bool]
            Whether or not category should be archived.
        group_id: Optional[int]
            Assigns the newly-created category to an existing category group.

        Returns
        -------
        int
            ID of the newly created category
        """
        category_body = ModelCreateCategory(
            name=name,
            description=description,
            is_income=is_income,
            exclude_from_budget=exclude_from_budget,
            exclude_from_totals=exclude_from_totals,
            archived=archived,
            group_id=group_id,
        ).model_dump(exclude_none=True)
        response_data = self.make_request(
            method=self.Methods.POST,
            url_path=APIConfig.LUNCHMONEY_CATEGORIES,
            payload=category_body,
        )
        return response_data["category_id"]

    def get_category(self, category_id: int) -> CategoriesObject:
        """
        Get single category

        Use this endpoint to get hydrated details on a single category. Note that if
        this category is part of a category group, its properties (is_income,
        exclude_from_budget, exclude_from_totals) will inherit from the category group.

        https://lunchmoney.dev/#get-single-category

        Parameters
        ----------
        category_id : int
            Id of the Lunch Money Category

        Returns
        -------
        CategoriesObject
        """
        response_data = self.make_request(
            method=self.Methods.GET,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, category_id],
        )
        return CategoriesObject.model_validate(response_data)

    def remove_category(self, category_id: int) -> bool:
        """
        Delete a single category

        Use this endpoint to delete a single category or category group. This will
        only work if there are no dependencies, such as existing budgets for the
        category, categorized transactions, categorized recurring items, etc. If
        there are dependents, this endpoint will return what the dependents are
        and how many there are.

        https://lunchmoney.dev/#delete-category

        Parameters
        ----------
        category_id : int
            Id of the Lunch Money Category

        Returns
        -------
        bool
        """
        response_data = self.make_request(
            method=self.Methods.DELETE,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, category_id],
        )
        if response_data is not True:
            raise LunchMoneyError(
                f"That Category ({category_id}) has Dependents: "
                f"{json.dumps(response_data, indent=4)}"
            )
        return response_data

    def remove_category_force(self, category_id: int) -> bool:
        """
        Forcefully delete a single category

        Use this endpoint to force delete a single category or category group and
        along with it, disassociate the category from any transactions, recurring
        items, budgets, etc.

        Note: it is best practice to first try the Delete Category endpoint to ensure
        you don't accidentally delete any data. Disassociation/deletion of the data
        arising from this endpoint is irreversible!

        https://lunchmoney.dev/#force-delete-category

        Parameters
        ----------
        category_id : int
            Id of the Lunch Money Category

        Returns
        -------
        bool
        """
        response_data = self.make_request(
            method=self.Methods.DELETE,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, category_id, "force"],
        )
        return response_data

    def update_category(
        self,
        category_id: int,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_income: Optional[bool] = None,
        exclude_from_budget: Optional[bool] = None,
        exclude_from_totals: Optional[bool] = None,
        group_id: Optional[int] = None,
        archived: Optional[bool] = None,
    ) -> bool:
        """
        Update a single category

        Use this endpoint to update the properties for a single category or category group

        https://lunchmoney.dev/#update-category

        Parameters
        ----------
        category_id : int
            Id of the Lunch Money Category
        name: str
            Name of category. Must be between 1 and 40 characters.
        description: Optional[str]
            Description of category. Must be less than 140 categories. Defaults to None.
        is_income: Optional[bool]
            Whether or not transactions in this category should be treated as income.
            Defaults to False.
        exclude_from_budget: Optional[bool]
            Whether or not transactions in this category should be excluded from budgets.
            Defaults to False.
        exclude_from_totals: Optional[bool]
            Whether or not transactions in this category should be excluded from
            calculated totals. Defaults to False.
        group_id: Optional[int]
            For a category, set the group_id to include it in a category group
        archived: Optional[bool]
            Whether or not category should be archived.

        Returns
        -------
        bool
        """
        payload = _CategoriesParamsPut(
            name=name,
            description=description,
            is_income=is_income,
            exclude_from_budget=exclude_from_budget,
            exclude_from_totals=exclude_from_totals,
            group_id=group_id,
            archived=archived,
        ).model_dump(exclude_none=True)
        response_data = self.make_request(
            method=self.Methods.PUT,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, category_id],
            payload=payload,
        )
        return response_data

    def insert_category_group(
        self,
        name: str,
        description: Optional[str] = None,
        is_income: Optional[bool] = None,
        exclude_from_budget: Optional[bool] = None,
        exclude_from_totals: Optional[bool] = None,
        category_ids: Optional[List[int]] = None,
        new_categories: Optional[List[str]] = None,
    ) -> int:
        """
        Create a Spending Category Group

        Use this endpoint to create a single category group
        https://lunchmoney.dev/#create-category-group

        Parameters
        ----------
        name: str
            Name of category. Must be between 1 and 40 characters.
        description: Optional[str]
            Description of category. Must be less than 140 categories. Defaults to None.
        is_income: Optional[bool]
            Whether or not transactions in this category should be treated as income.
            Defaults to False.
        exclude_from_budget: Optional[bool]
            Whether or not transactions in this category should be excluded from budgets.
            Defaults to False.
        exclude_from_totals: Optional[bool]
            Whether or not transactions in this category should be excluded from
            calculated totals. Defaults to False.
        category_ids: Optional[List[int]]
            Array of category_id to include in the category group.
        new_categories: Optional[List[str]]
            Array of strings representing new categories to create and subsequently
            include in the category group.

        Returns
        -------
        int
            ID of the newly created category group
        """
        payload = _CategoriesParamsPost(
            name=name,
            description=description,
            is_income=is_income,
            exclude_from_budget=exclude_from_budget,
            exclude_from_totals=exclude_from_totals,
            category_ids=category_ids,
            new_categories=new_categories,
        ).model_dump(exclude_none=True)
        response_data = self.make_request(
            method=self.Methods.POST,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, "group"],
            payload=payload,
        )
        return response_data["category_id"]

    def insert_into_category_group(
        self,
        category_group_id: int,
        category_ids: Optional[List[int]] = None,
        new_categories: Optional[List[str]] = None,
    ) -> CategoriesObject:
        """
        Add to a Category Group

        Use this endpoint to add categories (either existing or new) to a single
        category group

        https://lunchmoney.dev/#add-to-category-group

        Parameters
        ----------
        category_group_id: int
            Id of the Lunch Money Category Group
        category_ids: Optional[List[int]]
            Array of category_id to include in the category group.
        new_categories: Optional[List[str]]
            Array of strings representing new categories to create and subsequently
            include in the category group.

        Returns
        -------
        CategoriesObject
        """
        payload = _CategoriesAddParamsPost(
            category_ids=category_ids, new_categories=new_categories
        ).model_dump(exclude_none=True)
        response_data = self.make_request(
            method=self.Methods.POST,
            url_path=[
                APIConfig.LUNCHMONEY_CATEGORIES,
                "group",
                category_group_id,
                "add",
            ],
            payload=payload,
        )
        return CategoriesObject.model_validate(response_data)



"""
Lunch Money - User

https://lunchmoney.dev/#user
"""

import logging
from typing import Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient

logger = logging.getLogger(__name__)


class UserObject(LunchableModel):
    """
    The LunchMoney `User` object

    https://lunchmoney.dev/#user-object
    """

    user_id: int = Field(description="Unique identifier for user")
    user_name: str = Field(description="User's' name")
    user_email: str = Field(description="User's' Email")
    account_id: int = Field(
        description="Unique identifier for the associated budgeting account"
    )
    budget_name: str = Field(description="Name of the associated budgeting account")
    api_key_label: Optional[str] = Field(
        None,
        description="User-defined label of the developer API key used. "
        "Returns null if nothing has been set.",
    )


class UserClient(LunchMoneyAPIClient):
    """
    Lunch Money Interactions for Non Finance Operations
    """

    def get_user(self) -> UserObject:
        """
        Get Personal User Details

        Use this endpoint to get details on the current user.

        https://lunchmoney.dev/#get-user

        Returns
        -------
        UserObject
        """
        response_data = self.make_request(
            method=self.Methods.GET, url_path=APIConfig.LUNCHMONEY_ME
        )
        me = UserObject.model_validate(response_data)
        return me



"""
Run Tests on the Budgets Endpoint
"""

import datetime
import logging

from lunchable import LunchMoney
from lunchable.models import BudgetObject
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_upsert_budget(
    lunch_money_obj: LunchMoney, obscure_start_date: datetime.datetime
):
    """
    Test upserting some budgets
    """
    # Ride Sharing
    response = lunch_money_obj.upsert_budget(
        start_date=obscure_start_date, category_id=443127, amount=100.00
    )
    assert isinstance(response, dict) or response is None


@lunchable_cassette
def test_get_budgets(
    lunch_money_obj: LunchMoney, obscure_start_date: datetime.datetime
):
    """
    Test Getting some budgets
    """
    budgets = lunch_money_obj.get_budgets(
        start_date=obscure_start_date,
        end_date=obscure_start_date + datetime.timedelta(days=28),
    )
    assert len(budgets) >= 1
    for budget in budgets:
        assert isinstance(budget, BudgetObject)
    logger.info("%s Budgets Found", len(budgets))
    logger.info(budgets)


@lunchable_cassette
def test_delete_budget(
    lunch_money_obj: LunchMoney, obscure_start_date: datetime.datetime
):
    """
    Delete a Budget
    """
    deleted = lunch_money_obj.remove_budget(
        start_date=obscure_start_date, category_id=443127
    )
    logger.info("Budget Deleted")
    assert deleted is True



"""
Run Tests on the Plaid Accounts Endpoint
"""

import logging

from lunchable import LunchMoney
from lunchable.models.plaid_accounts import PlaidAccountObject
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_get_plaid_accounts(lunch_money_obj: LunchMoney):
    """
    Get Plaid Account and Assert it's a Plaid Account
    """
    plaid_accounts = lunch_money_obj.get_plaid_accounts()
    assert len(plaid_accounts) >= 1
    for plaid_account in plaid_accounts:
        assert isinstance(plaid_account, PlaidAccountObject)
    logger.info("%s Plaid Accounts returned", len(plaid_accounts))


@lunchable_cassette
def test_trigger_fetch_from_plaid(lunch_money_obj: LunchMoney):
    """
    Trigger Plaid Fetch
    """
    plaid_fetch_request = lunch_money_obj.trigger_fetch_from_plaid()
    assert plaid_fetch_request is True



"""
Lunch Money - Plaid Accounts

https://lunchmoney.dev/#plaid-accounts
"""

from __future__ import annotations

import datetime
import logging
from typing import List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient
from lunchable.models._descriptions import _PlaidAccountDescriptions

logger = logging.getLogger(__name__)


class _PlaidFetchRequest(LunchableModel):
    """
    Trigger Fetch from Plaid

    https://lunchmoney.dev/#trigger-fetch-from-plaid
    """

    start_date: Optional[datetime.date] = None
    end_date: Optional[datetime.date] = None
    plaid_account_id: Optional[int] = None


class PlaidAccountObject(LunchableModel):
    """
    Assets synced from Plaid

    Similar to AssetObjects, these accounts are linked to remote sources in Plaid.

    https://lunchmoney.dev/#plaid-accounts-object
    """

    id: int = Field(description="Unique identifier of Plaid account")
    date_linked: datetime.date = Field(
        description=_PlaidAccountDescriptions.date_linked
    )
    name: str = Field(description=_PlaidAccountDescriptions.name)
    type: str = Field(description=_PlaidAccountDescriptions.type)
    subtype: str = Field(description=_PlaidAccountDescriptions.subtype)
    mask: Optional[str] = Field(None, description=_PlaidAccountDescriptions.mask)
    institution_name: str = Field(
        description=_PlaidAccountDescriptions.institution_name
    )
    status: str = Field(description=_PlaidAccountDescriptions.status)
    last_import: Optional[datetime.datetime] = Field(
        None, description=_PlaidAccountDescriptions.last_import
    )
    balance: Optional[float] = Field(
        None, description=_PlaidAccountDescriptions.balance
    )
    currency: str = Field(description=_PlaidAccountDescriptions.currency)
    balance_last_update: datetime.datetime = Field(
        description=_PlaidAccountDescriptions.balance_last_update
    )
    limit: Optional[int] = Field(None, description=_PlaidAccountDescriptions.limit)


class PlaidAccountsClient(LunchMoneyAPIClient):
    """
    Lunch Money Plaid Accounts Interactions
    """

    def get_plaid_accounts(self) -> List[PlaidAccountObject]:
        """
        Get Plaid Synced Assets

        Get a list of all synced Plaid accounts associated with the user's account.

        Plaid Accounts are individual bank accounts that you have linked to Lunch Money via Plaid.
        You may link one bank but one bank might contain 4 accounts. Each of these
        accounts is a Plaid Account. (https://lunchmoney.dev/#plaid-accounts-object)

        Returns
        -------
        List[PlaidAccountObject]
        """
        response_data = self.make_request(
            method=self.Methods.GET, url_path=APIConfig.LUNCHMONEY_PLAID_ACCOUNTS
        )
        accounts = response_data.get(APIConfig.LUNCHMONEY_PLAID_ACCOUNTS)
        account_objects = [PlaidAccountObject.model_validate(item) for item in accounts]
        return account_objects

    def trigger_fetch_from_plaid(
        self,
        start_date: Optional[datetime.date] = None,
        end_date: Optional[datetime.date] = None,
        plaid_account_id: Optional[int] = None,
    ) -> bool:
        """
        Trigger Fetch from Plaid

        ** This is an experimental endpoint and parameters and/or response may change. **

        Use this endpoint to trigger a fetch for latest data from Plaid.

        Returns true if there were eligible Plaid accounts to trigger a fetch for. Eligible
        accounts are those who last_fetch value is over 1 minute ago. (Although the limit
        is every minute, please use this endpoint sparingly!)

        Note that fetching from Plaid is a background job. This endpoint simply queues up
        the job. You may track the plaid_last_successful_update, last_fetch and last_import
        properties to verify the results of the fetch.

        Parameters
        ----------
        start_date: Optional[datetime.date]
            Start date for fetch (ignored if end_date is null)
        end_date: Optional[datetime.date]
            End date for fetch (ignored if start_date is null)
        plaid_account_id: Optional[int]
            Specific ID of a plaid account to fetch. If left empty,
            endpoint will trigger a fetch for all eligible accounts

        Returns
        -------
        bool
            Returns true if there were eligible Plaid accounts to trigger a fetch for.
        """
        fetch_request = _PlaidFetchRequest(
            start_date=start_date, end_date=end_date, plaid_account_id=plaid_account_id
        )
        response: bool = self.make_request(
            method=self.Methods.POST,
            url_path=[APIConfig.LUNCHMONEY_PLAID_ACCOUNTS, "fetch"],
            data=fetch_request.model_dump(exclude_none=True),
        )
        return response



"""
Lunch Money - Transactions

https://lunchmoney.dev/#transactions
"""

import datetime
import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import pydantic_core
from pydantic import Field, field_validator

from lunchable import LunchMoneyError
from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient
from lunchable.models._descriptions import (
    _TransactionDescriptions,
    _TransactionInsertDescriptions,
    _TransactionSplitDescriptions,
    _TransactionUpdateDescriptions,
)
from lunchable.models.tags import TagsObject

logger = logging.getLogger(__name__)


class TransactionBaseObject(LunchableModel):
    """
    Base Model For All Transactions to Inherit From
    """

    pass


class TransactionInsertObject(TransactionBaseObject):
    """
    Object For Creating New Transactions

    https://lunchmoney.dev/#insert-transactions
    """

    class StatusEnum(str, Enum):
        """
        Status Options, must be "cleared" or "uncleared"
        """

        cleared = "cleared"
        uncleared = "uncleared"

    date: datetime.date = Field(description=_TransactionInsertDescriptions.date)
    amount: float = Field(description=_TransactionInsertDescriptions.amount)
    category_id: Optional[int] = Field(
        None, description=_TransactionInsertDescriptions.category_id
    )
    payee: Optional[str] = Field(None, description="Max 140 characters", max_length=140)
    currency: Optional[str] = Field(
        None, description=_TransactionInsertDescriptions.currency, max_length=3
    )
    asset_id: Optional[int] = Field(
        None, description=_TransactionInsertDescriptions.asset_id
    )
    recurring_id: Optional[int] = Field(
        None, description=_TransactionInsertDescriptions.recurring_id
    )
    notes: Optional[str] = Field(None, description="Max 350 characters", max_length=350)
    status: Optional[StatusEnum] = Field(
        None, description=_TransactionInsertDescriptions.status
    )
    external_id: Optional[str] = Field(
        None, description=_TransactionInsertDescriptions.external_id, max_length=75
    )
    tags: Optional[List[Union[str, int]]] = Field(
        None, description=_TransactionInsertDescriptions.tags
    )


class TransactionUpdateObject(TransactionBaseObject):
    """
    Object For Updating Existing Transactions

    https://lunchmoney.dev/#update-transaction
    """

    class StatusEnum(str, Enum):
        """
        Status Options, must be "cleared" or "uncleared"
        """

        cleared = "cleared"
        uncleared = "uncleared"

    date: Optional[datetime.date] = Field(
        None, description=_TransactionUpdateDescriptions.date
    )
    category_id: Optional[int] = Field(
        None, description=_TransactionUpdateDescriptions.category_id
    )
    payee: Optional[str] = Field(None, description="Max 140 characters", max_length=140)
    amount: Optional[float] = Field(
        None, description=_TransactionUpdateDescriptions.amount
    )
    currency: Optional[str] = Field(
        None, description=_TransactionUpdateDescriptions.currency
    )
    asset_id: Optional[int] = Field(
        None, description=_TransactionUpdateDescriptions.asset_id
    )
    recurring_id: Optional[int] = Field(
        None, description=_TransactionUpdateDescriptions.recurring_id
    )
    notes: Optional[str] = Field(None, description="Max 350 characters", max_length=350)
    status: Optional[StatusEnum] = Field(
        None, description=_TransactionUpdateDescriptions.status
    )
    external_id: Optional[str] = Field(
        None, description=_TransactionUpdateDescriptions.external_id
    )
    tags: Optional[List[Union[int, str]]] = Field(
        None, description=_TransactionUpdateDescriptions.tags
    )


class TransactionSplitObject(TransactionBaseObject):
    """
    Object for Splitting Transactions

    https://lunchmoney.dev/#split-object
    """

    date: datetime.date = Field(description=_TransactionSplitDescriptions.date)
    category_id: Optional[int] = Field(
        default=None, description=_TransactionSplitDescriptions.category_id
    )
    notes: Optional[str] = Field(None, description=_TransactionSplitDescriptions.notes)
    amount: float = Field(description=_TransactionSplitDescriptions.amount)


class FullStatusEnum(str, Enum):
    """
    Status Options
    """

    cleared = "cleared"
    uncleared = "uncleared"
    recurring = "recurring"
    recurring_suggested = "recurring_suggested"
    pending = "pending"


class TransactionChildObject(TransactionBaseObject):
    """
    Child Transaction Object for Transaction Groups
    """

    id: int = Field(description=_TransactionDescriptions.id)
    amount: float = Field(description=_TransactionDescriptions.amount)
    payee: Optional[str] = Field(None, description=_TransactionDescriptions.payee)
    date: datetime.date = Field(description=_TransactionDescriptions.date)
    formatted_date: Optional[str] = Field(
        None, description=_TransactionDescriptions.formatted_date
    )
    notes: Optional[str] = Field(None, description=_TransactionDescriptions.notes)
    currency: Optional[str] = Field(None, description=_TransactionDescriptions.currency)
    asset_id: Optional[int] = Field(None, description=_TransactionDescriptions.asset_id)
    plaid_account_id: Optional[int] = Field(
        None, description=_TransactionDescriptions.plaid_account_id
    )
    to_base: Optional[float] = Field(None, description=_TransactionDescriptions.to_base)


class TransactionObject(TransactionBaseObject):
    """
    Universal Lunch Money Transaction Object

    https://lunchmoney.dev/#transaction-object
    """

    id: int = Field(description=_TransactionDescriptions.id)
    date: datetime.date = Field(description=_TransactionDescriptions.date)
    payee: Optional[str] = Field(None, description=_TransactionDescriptions.payee)
    amount: float = Field(description=_TransactionDescriptions.amount)
    currency: Optional[str] = Field(
        None, max_length=3, description=_TransactionDescriptions.currency
    )
    to_base: Optional[float] = Field(None, description=_TransactionDescriptions.to_base)
    category_id: Optional[int] = Field(
        None, description=_TransactionDescriptions.category_id
    )
    category_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.category_name
    )
    category_group_id: Optional[int] = Field(
        None, description=_TransactionDescriptions.category_group_id
    )
    category_group_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.category_group_name
    )
    is_income: Optional[bool] = Field(
        None, description=_TransactionDescriptions.is_income
    )
    exclude_from_budget: Optional[bool] = Field(
        None, description=_TransactionDescriptions.exclude_from_budget
    )
    exclude_from_totals: Optional[bool] = Field(
        None, description=_TransactionDescriptions.exclude_from_totals
    )
    created_at: datetime.datetime = Field(
        description=_TransactionDescriptions.created_at
    )
    updated_at: datetime.datetime = Field(
        description=_TransactionDescriptions.updated_at
    )
    status: Optional[str] = Field(None, description=_TransactionDescriptions.status)
    is_pending: Optional[bool] = Field(
        None, description=_TransactionDescriptions.is_pending
    )
    notes: Optional[str] = Field(None, description=_TransactionDescriptions.notes)
    original_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.original_name
    )
    recurring_id: Optional[int] = Field(
        None, description=_TransactionDescriptions.recurring_id
    )
    recurring_payee: Optional[str] = Field(
        None, description=_TransactionDescriptions.recurring_payee
    )
    recurring_description: Optional[str] = Field(
        None, description=_TransactionDescriptions.recurring_description
    )
    recurring_cadence: Optional[str] = Field(
        None, description=_TransactionDescriptions.recurring_cadence
    )
    recurring_type: Optional[str] = Field(
        None, description=_TransactionDescriptions.recurring_type
    )
    recurring_amount: Optional[float] = Field(
        None, description=_TransactionDescriptions.recurring_amount
    )
    recurring_currency: Optional[str] = Field(
        None, description=_TransactionDescriptions.recurring_currency
    )
    parent_id: Optional[int] = Field(
        None, description=_TransactionDescriptions.parent_id
    )
    has_children: Optional[bool] = Field(
        None, description=_TransactionDescriptions.has_children
    )
    group_id: Optional[int] = Field(None, description=_TransactionDescriptions.group_id)
    is_group: Optional[bool] = Field(
        None, description=_TransactionDescriptions.is_group
    )
    asset_id: Optional[int] = Field(None, description=_TransactionDescriptions.asset_id)
    asset_institution_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.asset_institution_name
    )
    asset_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.asset_name
    )
    asset_display_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.asset_display_name
    )
    asset_status: Optional[str] = Field(
        None, description=_TransactionDescriptions.asset_status
    )
    plaid_account_id: Optional[int] = Field(
        None, description=_TransactionDescriptions.plaid_account_id
    )
    plaid_account_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.plaid_account_name
    )
    plaid_account_mask: Optional[str] = Field(
        None, description=_TransactionDescriptions.plaid_account_mask
    )
    institution_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.institution_name
    )
    plaid_account_display_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.plaid_account_display_name
    )
    plaid_metadata: Optional[Dict[str, Any]] = Field(
        None, description=_TransactionDescriptions.plaid_metadata
    )
    source: Optional[str] = Field(None, description=_TransactionDescriptions.source)
    display_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.display_name
    )
    display_notes: Optional[str] = Field(
        None, description=_TransactionDescriptions.display_notes
    )
    account_display_name: Optional[str] = Field(
        None, description=_TransactionDescriptions.account_display_name
    )
    tags: Optional[List[TagsObject]] = Field(None, description="Array of Tag objects")
    external_id: Optional[str] = Field(
        None, max_length=75, description=_TransactionDescriptions.external_id
    )
    children: Optional[List[TransactionChildObject]] = Field(
        None, description=_TransactionDescriptions.children
    )

    @field_validator("plaid_metadata", mode="before")
    def to_json(cls, x: Optional[str]) -> Optional[Dict[str, Any]]:
        """
        Check a result
        """
        if x is None:
            return None
        elif isinstance(x, dict):
            return x
        else:
            return pydantic_core.from_json(x)

    def get_update_object(self) -> TransactionUpdateObject:
        """
        Return a TransactionUpdateObject

        Return a TransactionUpdateObject to update an expense. Simply
        change one of the properties and perform an `update_transaction` with
        your Lunchable object.

        Returns
        -------
        TransactionUpdateObject
        """
        update_dict = self.model_dump()
        try:
            TransactionUpdateObject.StatusEnum(self.status)
        except ValueError:
            update_dict["status"] = None
        update_object = TransactionUpdateObject.model_validate(update_dict)
        if update_object.tags is not None:
            tags = [] if self.tags is None else self.tags
            update_object.tags = [tag.name for tag in tags]
        return update_object

    def get_insert_object(self) -> TransactionInsertObject:
        """
        Return a TransactionInsertObject

        Return a TransactionInsertObject to update an expense. Simply
        change some of the properties and perform an `insert_transactions` with
        your Lunchable object.

        Returns
        -------
        TransactionInsertObject
        """
        insert_dict = self.model_dump()
        try:
            TransactionInsertObject.StatusEnum(self.status)
        except ValueError:
            insert_dict["status"] = None
        insert_object = TransactionInsertObject.model_validate(insert_dict)
        if insert_object.tags is not None:
            tags = [] if self.tags is None else self.tags
            insert_object.tags = [tag.name for tag in tags]
        return insert_object


class _TransactionsResponse(LunchableModel):
    """
    HTTP Response for Transactions
    """

    transactions: List[TransactionObject]
    has_more: bool = False


class _TransactionParamsGet(LunchableModel):
    """
    https://lunchmoney.dev/#get-all-transactions
    """

    tag_id: Optional[int] = None
    recurring_id: Optional[int] = None
    plaid_account_id: Optional[int] = None
    category_id: Optional[int] = None
    asset_id: Optional[int] = None
    group_id: Optional[int] = None
    is_group: Optional[bool] = None
    status: Optional[FullStatusEnum] = None
    offset: Optional[int] = None
    limit: Optional[int] = None
    start_date: Optional[datetime.date] = None
    end_date: Optional[datetime.date] = None
    debit_as_negative: Optional[bool] = None
    pending: Optional[bool] = None


class _TransactionInsertParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#insert-transactions
    """

    transactions: List[TransactionInsertObject]
    apply_rules: Optional[bool] = None
    skip_duplicates: Optional[bool] = None
    check_for_recurring: Optional[bool] = None
    debit_as_negative: Optional[bool] = None
    skip_balance_update: Optional[bool] = None


class _TransactionGroupParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#create-transaction-group
    """

    date: datetime.date
    payee: str
    category_id: Optional[int] = None
    notes: Optional[str] = None
    tags: Optional[List[int]] = None
    transactions: List[int]


class _TransactionUpdateParamsPut(LunchableModel):
    """
    https://lunchmoney.dev/#update-transaction
    """

    split: Optional[List[TransactionSplitObject]] = None
    transaction: Optional[TransactionUpdateObject] = None
    debit_as_negative: Optional[bool] = None
    skip_balance_update: Optional[bool] = None


class _TransactionsUnsplitPost(LunchableModel):
    """
    https://lunchmoney.dev/#unsplit-transactions
    """

    parent_ids: List[int]
    remove_parents: Optional[bool] = None


class TransactionsClient(LunchMoneyAPIClient):
    """
    Lunch Money Transactions Interactions
    """

    def get_transactions(
        self,
        start_date: Optional[Union[datetime.date, datetime.datetime, str]] = None,
        end_date: Optional[Union[datetime.date, datetime.datetime, str]] = None,
        tag_id: Optional[int] = None,
        recurring_id: Optional[int] = None,
        plaid_account_id: Optional[int] = None,
        category_id: Optional[int] = None,
        asset_id: Optional[int] = None,
        group_id: Optional[int] = None,
        is_group: Optional[bool] = None,
        status: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        debit_as_negative: Optional[bool] = None,
        pending: Optional[bool] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[TransactionObject]:
        """
        Get Transactions Using Criteria

        Use this to retrieve all transactions between a date range (this method handles
        pagination automatically unless you specify a limit / offset argument). Returns
        list of Transaction objects. If no query parameters are set, this will return
        transactions for the current calendar month. If either start_date or end_date are
        datetime.datetime objects, they will be reduced to dates. If a string is provided,
        it will be attempted to be parsed as YYYY-MM-DD format.

        Parameters
        ----------
        start_date: Optional[Union[datetime.date, datetime.datetime, str]]
            Denotes the beginning of the time period to fetch transactions for. Defaults
            to beginning of current month. Required if end_date exists. Format: YYYY-MM-DD.
        end_date: Optional[Union[datetime.date, datetime.datetime, str]]
            Denotes the end of the time period you'd like to get transactions for.
            Defaults to end of current month. Required if start_date exists.
        tag_id: Optional[int]
            Filter by tag. Only accepts IDs, not names.
        recurring_id: Optional[int]
            Filter by recurring expense
        plaid_account_id: Optional[int]
            Filter by Plaid account
        category_id: Optional[int]
            Filter by category. Will also match category groups.
        asset_id: Optional[int]
            Filter by asset
        group_id: Optional[int]
            Filter by group_id (if the transaction is part of a specific group)
        is_group: Optional[bool]
            Filter by group (returns transaction groups)
        status: Optional[str]
            Filter by status (Can be cleared or uncleared. For recurring
            transactions, use recurring)
        offset: Optional[int]
            Sets the offset for the records returned (disables
            automatic pagination)
        limit: Optional[int]
            Sets the maximum number of records to return. Defaults to 1000
             (disables automatic pagination)
        debit_as_negative: Optional[bool]
            Pass in true if you'd like expenses to be returned as negative amounts and
            credits as positive amounts. Defaults to false.
        pending: Optional[bool]
            Pass in true if you'd like to include imported transactions with a pending status.
        params: Optional[dict]
            Additional Query String Params

        Returns
        -------
        List[TransactionObject]
            A list of transactions

        Examples
        --------
        Retrieve a list of
        [TransactionObject][lunchable.models.transactions.TransactionObject]

        ```python
        from lunchable import LunchMoney

        lunch = LunchMoney(access_token="xxxxxxx")
        transactions = lunch.get_transactions(start_date="2020-01-01",
                                              end_date="2020-01-31")
        ```
        """
        search_params = _TransactionParamsGet(
            tag_id=tag_id,
            recurring_id=recurring_id,
            plaid_account_id=plaid_account_id,
            category_id=category_id,
            asset_id=asset_id,
            group_id=group_id,
            is_group=is_group,
            status=status,
            offset=offset,
            limit=limit,
            start_date=start_date,
            end_date=end_date,
            debit_as_negative=debit_as_negative,
            pending=pending,
        ).model_dump(exclude_none=True)
        search_params.update(params if params is not None else {})
        auto_paginate = all(
            [
                offset is None,
                search_params.get("offset") is None,
                limit is None,
                search_params.get("limit") is None,
            ]
        )
        transactions = self._get_transactions(
            search_params=search_params,
            paginate=auto_paginate,
        )
        return transactions

    def _get_transactions(
        self,
        search_params: Dict[str, Any],
        existing_transactions: Optional[List[TransactionObject]] = None,
        paginate: bool = True,
    ) -> List[TransactionObject]:
        """
        Paginate Transactions
        """
        existing_transactions = existing_transactions or []
        transaction_response = self.make_request(
            method=self.Methods.GET,
            url_path=APIConfig.LUNCHMONEY_TRANSACTIONS,
            params=search_params,
        )
        transaction_response = _TransactionsResponse.model_validate(
            transaction_response
        )
        existing_transactions.extend(transaction_response.transactions)
        if transaction_response.has_more and paginate:
            search_params["offset"] = len(existing_transactions)
            return self._get_transactions(
                existing_transactions=existing_transactions,
                search_params=search_params,
                paginate=True,
            )
        return existing_transactions

    def get_transaction(
        self, transaction_id: int, debit_as_negative: Optional[bool] = None
    ) -> TransactionObject:
        """
        Get a Transaction by ID

        Parameters
        ----------
        transaction_id: int
            Lunch Money Transaction ID
        debit_as_negative: Optional[bool]
            Pass in true if you'd like expenses to be returned as negative
            amounts and credits as positive amounts. Defaults to false.

        Returns
        -------
        TransactionObject

        Examples
        --------
        Retrieve a single transaction by its ID

        ```python
        from lunchable import LunchMoney

        lunch = LunchMoney(access_token="xxxxxxx")
        transaction = lunch.get_transaction(transaction_id=1234)
        ```

        The above code returns a
        [TransactionObject][lunchable.models.transactions.TransactionObject]
        with ID # 1234 (assuming it exists)
        """
        response_data = self.make_request(
            method=self.Methods.GET,
            url_path=[APIConfig.LUNCHMONEY_TRANSACTIONS, transaction_id],
            params={"debit_as_negative": debit_as_negative}
            if debit_as_negative is not None
            else {},
        )
        return TransactionObject.model_validate(response_data)

    ListOrSingleTransactionUpdateObject = Optional[
        Union[TransactionUpdateObject, TransactionObject]
    ]

    ListOrSingleTransactionInsertObject = Union[
        TransactionObject,
        TransactionInsertObject,
        List[TransactionObject],
        List[TransactionInsertObject],
        List[Union[TransactionObject, TransactionInsertObject]],
    ]

    def update_transaction(
        self,
        transaction_id: int,
        transaction: ListOrSingleTransactionUpdateObject = None,
        split: Optional[List[TransactionSplitObject]] = None,
        debit_as_negative: Optional[bool] = None,
        skip_balance_update: Optional[bool] = None,
    ) -> Dict[str, Any]:
        """
        Update a Transaction

        Use this endpoint to update a single transaction. You may also use this
        to split an existing transaction. If a TransactionObject is provided it will be
        converted into a TransactionUpdateObject.

        PUT https://dev.lunchmoney.app/v1/transactions/:transaction_id

        Parameters
        ----------
        transaction_id: int
            Lunch Money Transaction ID
        transaction: ListOrSingleTransactionUpdateObject
            Object to update with
        split: Optional[List[TransactionSplitObject]]
            Defines the split of a transaction. You may not split an already-split
            transaction, recurring transaction, or group transaction.
        debit_as_negative: Optional[bool]
            If true, will assume negative amount values denote expenses and
            positive amount values denote credits. Defaults to false.
        skip_balance_update: Optional[bool]
            If false, will skip updating balance if an asset_id
            is present for any of the transactions.

        Returns
        -------
        Dict[str, Any]

        Examples
        --------
        Update a transaction with a
        [TransactionUpdateObject][lunchable.models.transactions.TransactionUpdateObject]

        ```python
        from datetime import datetime

        from lunchable import LunchMoney, TransactionUpdateObject

        lunch = LunchMoney(access_token="xxxxxxx")
        transaction_note = f"Updated on {datetime.now()}"
        notes_update = TransactionUpdateObject(notes=transaction_note)
        response = lunch.update_transaction(transaction_id=1234,
                                            transaction=notes_update)
        ```

        Update a
        [TransactionObject][lunchable.models.transactions.TransactionObject]
        with itself

        ```python
        from datetime import datetime, timedelta

        from lunchable import LunchMoney

        lunch = LunchMoney(access_token="xxxxxxx")
        transaction = lunch.get_transaction(transaction_id=1234)

        transaction.notes = f"Updated on {datetime.now()}"
        transaction.date = transaction.date + timedelta(days=1)
        response = lunch.update_transaction(transaction_id=transaction.id,
                                            transaction=transaction)
        ```
        """
        payload = _TransactionUpdateParamsPut(
            split=split,
            debit_as_negative=debit_as_negative,
            skip_balance_update=skip_balance_update,
        ).model_dump(exclude_none=True)
        if transaction is None and split is None:
            raise LunchMoneyError("You must update the transaction or provide a split")
        elif transaction is not None:
            if isinstance(transaction, TransactionObject):
                transaction = transaction.get_update_object()
            payload["transaction"] = transaction.model_dump(exclude_unset=True)
        response_data = self.make_request(
            method=self.Methods.PUT,
            url_path=[APIConfig.LUNCHMONEY_TRANSACTIONS, transaction_id],
            payload=payload,
        )
        return response_data

    def insert_transactions(
        self,
        transactions: ListOrSingleTransactionInsertObject,
        apply_rules: Optional[bool] = None,
        skip_duplicates: Optional[bool] = None,
        debit_as_negative: Optional[bool] = None,
        check_for_recurring: Optional[bool] = None,
        skip_balance_update: Optional[bool] = None,
    ) -> List[int]:
        """
        Create One or Many Lunch Money Transactions

        Use this endpoint to insert many transactions at once. Also accepts
        a single transaction as well. If a TransactionObject is provided it will be
        converted into a TransactionInsertObject.

        https://lunchmoney.dev/#insert-transactions

        Parameters
        ----------
        transactions: ListOrSingleTransactionTypeObject
            Transactions to insert. Either a single TransactionInsertObject object or
            a list of them
        apply_rules: Optional[bool]
            If true, will apply account's existing rules to the inserted transactions.
            Defaults to false.
        skip_duplicates: Optional[bool]
            If true, the system will automatically dedupe based on transaction date,
            payee and amount. Note that deduping by external_id will occur regardless
            of this flag.
        check_for_recurring: Optional[bool]
            if true, will check new transactions for occurrences of new monthly expenses.
            Defaults to false.
        debit_as_negative: Optional[bool]
            If true, will assume negative amount values denote expenses and
            positive amount values denote credits. Defaults to false.
        skip_balance_update: Optional[bool]
            If false, will skip updating balance if an asset_id
            is present for any of the transactions.

        Returns
        -------
        List[int]

        Examples
        --------
        Create a new transaction with a
        [TransactionInsertObject][lunchable.models.transactions.TransactionInsertObject]

        ```python
        from lunchable import LunchMoney, TransactionInsertObject

        lunch = LunchMoney(access_token="xxxxxxx")

        new_transaction = TransactionInsertObject(payee="Example Restaurant",
                                                  amount=120.00,
                                                  notes="Saturday Dinner")
        new_transaction_ids = lunch.insert_transactions(transactions=new_transaction)
        ```
        """
        insert_objects = []
        if not isinstance(transactions, list):
            transactions = [transactions]
        for item in transactions:
            if isinstance(item, TransactionObject):
                insert_objects.append(item.get_insert_object())
            elif isinstance(item, TransactionInsertObject):
                insert_objects.append(item)
            else:
                raise LunchMoneyError(
                    "Only TransactionObjects or TransactionInsertObjects are "
                    "supported by this function."
                )
        payload = _TransactionInsertParamsPost(
            transactions=insert_objects,
            apply_rules=apply_rules,
            skip_duplicates=skip_duplicates,
            check_for_recurring=check_for_recurring,
            debit_as_negative=debit_as_negative,
            skip_balance_update=skip_balance_update,
        ).model_dump(exclude_none=True)
        response_data = self.make_request(
            method=self.Methods.POST,
            url_path=APIConfig.LUNCHMONEY_TRANSACTIONS,
            payload=payload,
        )
        ids: List[int] = response_data["ids"] if response_data else []
        return ids

    def insert_transaction_group(
        self,
        date: datetime.date,
        payee: str,
        transactions: List[int],
        category_id: Optional[int] = None,
        notes: Optional[str] = None,
        tags: Optional[List[int]] = None,
    ) -> int:
        """
        Create a Transaction Group of Two or More Transactions

        Returns the ID of the newly created transaction group

        Parameters
        ----------
        date: datetime.date
            Date for the grouped transaction
        payee: str
            Payee name for the grouped transaction
        category_id: Optional[int]
            Category for the grouped transaction
        notes: Optional[str]
            Notes for the grouped transaction
        tags: Optional[List[int]]
            Array of tag IDs for the grouped transaction
        transactions: Optional[List[int]]
            Array of transaction IDs to be part of the transaction group

        Returns
        -------
        int
        """
        if len(transactions) < 2:
            raise LunchMoneyError(
                "You must include 2 or more transactions " "in the Transaction Group"
            )
        transaction_params = _TransactionGroupParamsPost(
            date=date,
            payee=payee,
            category_id=category_id,
            notes=notes,
            tags=tags,
            transactions=transactions,
        ).model_dump(exclude_none=True)
        response_data = self.make_request(
            method=self.Methods.POST,
            url_path=[
                APIConfig.LUNCHMONEY_TRANSACTIONS,
                APIConfig.LUNCHMONEY_TRANSACTION_GROUPS,
            ],
            payload=transaction_params,
        )
        return response_data

    def remove_transaction_group(self, transaction_group_id: int) -> List[int]:
        """
        Delete a Transaction Group

        Use this method to delete a transaction group. The transactions within the
        group will not be removed.

        Returns the IDs of the transactions that were part of the deleted group

        https://lunchmoney.dev/#delete-transaction-group

        Parameters
        ----------
        transaction_group_id: int
            Transaction Group Identifier

        Returns
        -------
        List[int]
        """
        response_data = self.make_request(
            method=self.Methods.DELETE,
            url_path=[
                APIConfig.LUNCHMONEY_TRANSACTIONS,
                APIConfig.LUNCHMONEY_TRANSACTION_GROUPS,
                transaction_group_id,
            ],
        )
        return response_data["transactions"]

    def unsplit_transactions(
        self, parent_ids: List[int], remove_parents: Optional[bool] = None
    ) -> List[int]:
        """
        Unsplit Transactions

        Use this endpoint to unsplit one or more transactions.

        Returns an array of IDs of deleted transactions

        https://lunchmoney.dev/#unsplit-transactions

        Parameters
        ----------
        parent_ids: List[int]
            Array of transaction IDs to unsplit. If one transaction is unsplittable,
            no transaction will be unsplit.
        remove_parents: Optional[bool]
            If true, deletes the original parent transaction as well. Note,
            this is unreversable!

        Returns
        -------
        List[int]
        """
        response_data = self.make_request(
            method=self.Methods.POST,
            url_path=[APIConfig.LUNCHMONEY_TRANSACTIONS, "unsplit"],
            payload=_TransactionsUnsplitPost(
                parent_ids=parent_ids, remove_parents=remove_parents
            ).model_dump(exclude_none=True),
        )
        return response_data

    def get_transaction_group(self, transaction_id: int) -> TransactionObject:
        """
        Get a Transaction Group

        Parameters
        ----------
        transaction_id: int
            Transaction ID of either the parent or any of the children
            in the transaction group

        Returns
        -------
        TransactionObject
            The transaction group as a `TransactionObject`
        """
        response_data = self.make_request(
            method=self.Methods.GET,
            params={"transaction_id": transaction_id},
            url_path=[
                APIConfig.LUNCHMONEY_TRANSACTIONS,
                APIConfig.LUNCHMONEY_TRANSACTION_GROUPS,
            ],
        )
        return TransactionObject.model_validate(response_data)



"""
Run Tests on the Categories Endpoint
"""

import logging

from lunchable import LunchMoney
from lunchable.models.categories import CategoriesObject
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_get_categories(lunch_money_obj: LunchMoney):
    """
    Get Categories and Assert that they're categories
    """
    categories = lunch_money_obj.get_categories()
    assert len(categories) >= 1
    for category in categories:
        assert isinstance(category, CategoriesObject)
    logger.info("%s Categories returned", len(categories))


@lunchable_cassette
def test_create_category(lunch_money_obj: LunchMoney):
    """
    Create A Single Category
    """
    name = "Test Category"
    category = lunch_money_obj.insert_category(
        name=name, description="Test Category Description", exclude_from_budget=True
    )
    logger.info("Category ID # %s was just created: %s", category, name)
    assert isinstance(category, int)


@lunchable_cassette
def test_get_category(lunch_money_obj: LunchMoney):
    """
    Get a Single Category
    """
    category_id = 443128
    category = lunch_money_obj.get_category(category_id=category_id)
    logger.info("Category ID # %s was just fetched: %s", category.id, category.name)
    assert isinstance(category, CategoriesObject)


@lunchable_cassette
def test_delete_category(lunch_money_obj: LunchMoney):
    """
    Delete a Category
    """
    category_id = 343088
    deleted = lunch_money_obj.remove_category(category_id=category_id)
    logger.info("Category ID # %s was just deleted", category_id)
    assert deleted is True


@lunchable_cassette
def test_delete_category_force(lunch_money_obj: LunchMoney):
    """
    Forcefully Delete a Category
    """
    category_id = 343089
    deleted = lunch_money_obj.remove_category_force(category_id=category_id)
    logger.info("Category ID # %s was just deleted", category_id)
    assert deleted is True


@lunchable_cassette
def test_update_category(lunch_money_obj: LunchMoney):
    """
    Update a Single Category
    """
    category_id = 443128
    updated = lunch_money_obj.update_category(
        category_id=category_id, description="Test Category Description Updated"
    )
    assert isinstance(updated, bool)


@lunchable_cassette
def test_create_category_group(lunch_money_obj: LunchMoney):
    """
    Create A Single Category Group
    """
    name = "Test Category Group"
    category_id = lunch_money_obj.insert_category_group(
        name=name, description="Test Category Group!!", exclude_from_budget=True
    )
    logger.info("Category Group ID # %s was just created: %s", category_id, name)
    assert isinstance(category_id, int)


@lunchable_cassette
def test_add_to_category_group(lunch_money_obj: LunchMoney):
    """
    Create A Single Category Group
    """
    name = "Test Category Group"
    category = lunch_money_obj.insert_into_category_group(
        category_group_id=658694,
        new_categories=["Another Another Test Category"],
        category_ids=[443128],
    )
    logger.info("Category Group ID # %s was just created: %s", category.id, name)
    assert isinstance(category, CategoriesObject)


@lunchable_cassette
def test_get_categories_nested(lunch_money_obj: LunchMoney):
    """
    Get Categories and Assert that they're categories
    """
    categories = lunch_money_obj.get_categories(format="nested")
    categories_with_children = list(filter(lambda x: x.children, categories))
    assert len(categories_with_children) >= 1


@lunchable_cassette
def test_get_categories_flattened(lunch_money_obj: LunchMoney):
    """
    Get Categories and Assert that they're categories
    """
    categories = lunch_money_obj.get_categories(format="flattened")
    assert len(categories) >= 1
    for category in categories:
        assert isinstance(category, CategoriesObject)



"""
Lunchmoney CLI
"""

import logging
import sys
from json import JSONDecodeError
from typing import Any, Dict, Optional

import click
import httpx
from click_plugins import with_plugins
from importlib_metadata import entry_points
from pydantic_core import to_jsonable_python
from rich import print, print_json, traceback

import lunchable
from lunchable import LunchMoney
from lunchable._config.logging_config import set_up_logging
from lunchable.models import LunchableModel

logger = logging.getLogger(__name__)


class LunchMoneyContext(LunchableModel):
    """
    Context Object to PAss Around CLI
    """

    debug: bool
    access_token: Optional[str]


debug_option = click.option(
    "--debug/--no-debug", default=False, help="Enable extra debugging output"
)
access_token_option = click.option(
    "--access-token",
    default=None,
    help="LunchMoney Developer API Access Token",
    envvar="LUNCHMONEY_ACCESS_TOKEN",
)


@click.group(invoke_without_command=True)
@click.version_option(
    version=lunchable.__version__, prog_name=lunchable.__application__
)
@access_token_option
@debug_option
@click.pass_context
def cli(ctx: click.core.Context, debug: bool, access_token: str) -> None:
    """
    Interactions with Lunch Money via lunchable 🍱
    """
    ctx.obj = LunchMoneyContext(debug=debug, access_token=access_token)
    traceback.install(show_locals=debug)
    set_up_logging(log_level=logging.DEBUG if debug is True else logging.INFO)
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@cli.group()
def transactions() -> None:
    """
    Interact with Lunch Money transactions
    """


@cli.group()
def plugins() -> None:
    """
    Interact with Lunchable Plugins

    Install lunchable with the "plugins" extra to get
    all the known plugins

    pipx install "lunchable[plugins]"
    """


@transactions.command("get")
@click.option(
    "--start-date",
    default=None,
    help="Denotes the beginning of the time period to fetch transactions for. Defaults"
    "to beginning of current month. Required if end_date exists. "
    "Format: YYYY-MM-DD.",
)
@click.option(
    "--end-date",
    default=None,
    help="Denotes the end of the time period you'd like to get transactions for. "
    "Defaults to end of current month. Required if start_date exists."
    "Format: YYYY-MM-DD.",
)
@click.option(
    "--tag-id", default=None, help="Filter by tag. Only accepts IDs, not names."
)
@click.option("--recurring-id", default=None, help="Filter by recurring expense")
@click.option("--plaid-account-id", default=None, help="Filter by Plaid account")
@click.option(
    "--category-id",
    default=None,
    help="Filter by category. Will also match category groups.",
)
@click.option("--asset-id", default=None, help="Filter by asset")
@click.option(
    "--group-id",
    default=None,
    help="Filter by group_id (if the transaction is part of a specific group)",
)
@click.option(
    "--is-group", default=None, help="Filter by group (returns transaction groups)"
)
@click.option(
    "--status",
    default=None,
    help="Filter by status (Can be cleared or uncleared. For recurring "
    "transactions, use recurring)",
)
@click.option("--offset", default=None, help="Sets the offset for the records returned")
@click.option(
    "--limit",
    default=None,
    help="Sets the maximum number of records to return. Note: The server will not "
    "respond with any indication that there are more records to be returned. "
    "Please check the response length to determine if you should make another "
    "call with an offset to fetch more transactions.",
)
@click.option(
    "--debit-as-negative",
    default=None,
    help="Pass in true if you’d like expenses to be returned as negative amounts and "
    "credits as positive amounts. Defaults to false.",
)
@click.option(
    "--pending",
    is_flag=True,
    default=None,
    help="Pass in true if you’d like to include imported transactions with a pending status.",
)
@click.pass_obj
def lunchmoney_transactions(
    context: LunchMoneyContext, **kwargs: Dict[str, Any]
) -> None:
    """
    Retrieve Lunch Money Transactions
    """
    lunch = LunchMoney(access_token=context.access_token)
    transactions = lunch.get_transactions(**kwargs)  # type: ignore[arg-type]
    json_data = to_jsonable_python(transactions)
    print_json(data=json_data)


@cli.command()
@click.argument("URL")
@click.option("-X", "--request", default="GET", help="Specify request command to use")
@click.option("-d", "--data", default=None, help="HTTP POST data")
@click.pass_obj
def http(context: LunchMoneyContext, url: str, request: str, data: str) -> None:
    """
    Interact with the LunchMoney API

    lunchable http /v1/transactions
    """
    lunch = LunchMoney(access_token=context.access_token)
    if not url.startswith("http"):
        url = url.lstrip("/")
        url_request = f"https://dev.lunchmoney.app/{url}"
    else:
        url_request = url
    resp = lunch.request(
        method=request,
        url=url_request,
        content=data,
    )
    try:
        resp.raise_for_status()
    except httpx.HTTPError:
        logger.error(resp)
        print(resp.text)
        sys.exit(1)
    try:
        response = resp.json()
    except JSONDecodeError:
        response = resp.text
    json_data = to_jsonable_python(response)
    print_json(data=json_data)


discovered_plugins = entry_points(group="lunchable.cli")
with_plugins(discovered_plugins)(plugins)



"""
Lunch Money - Recurring Expenses

https://lunchmoney.dev/#recurring-expenses
"""

from __future__ import annotations

import datetime
import logging
from typing import Any, Dict, List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient
from lunchable.models._descriptions import (
    _RecurringItemsDescriptions,
    _SummarizedTransactionDescriptions,
)

logger = logging.getLogger(__name__)


class SummarizedTransactionObject(LunchableModel):
    """
    Summarized Transaction Object
    """

    id: int = Field(description=_SummarizedTransactionDescriptions.id)
    date: datetime.date = Field(description=_SummarizedTransactionDescriptions.date)
    amount: float = Field(description=_SummarizedTransactionDescriptions.amount)
    currency: str = Field(
        max_length=3, description=_SummarizedTransactionDescriptions.currency
    )
    payee: str = Field(description=_SummarizedTransactionDescriptions.payee)
    category_id: Optional[int] = Field(
        None, description=_SummarizedTransactionDescriptions.category_id
    )
    recurring_id: Optional[int] = Field(
        None, description=_SummarizedTransactionDescriptions.recurring_id
    )
    to_base: float = Field(description=_SummarizedTransactionDescriptions.to_base)


class RecurringItemsObject(LunchableModel):
    """
    Recurring Expenses Object
    """

    id: int = Field(description=_RecurringItemsDescriptions.id)
    start_date: Optional[datetime.date] = Field(
        None, description=_RecurringItemsDescriptions.start_date
    )
    end_date: Optional[datetime.date] = Field(
        None, description=_RecurringItemsDescriptions.end_date
    )
    payee: str = Field(description=_RecurringItemsDescriptions.payee)
    currency: str = Field(
        max_length=3, description=_RecurringItemsDescriptions.currency
    )
    created_by: int = Field(description=_RecurringItemsDescriptions.created_by)
    created_at: datetime.datetime = Field(
        description=_RecurringItemsDescriptions.created_at
    )
    updated_at: datetime.datetime = Field(
        description=_RecurringItemsDescriptions.updated_at
    )
    billing_date: datetime.date = Field(
        description=_RecurringItemsDescriptions.billing_date
    )
    original_name: Optional[str] = Field(
        None, description=_RecurringItemsDescriptions.original_name
    )
    description: Optional[str] = Field(
        None, description=_RecurringItemsDescriptions.description
    )
    plaid_account_id: Optional[int] = Field(
        None, description=_RecurringItemsDescriptions.plaid_account_id
    )
    asset_id: Optional[int] = Field(
        None, description=_RecurringItemsDescriptions.asset_id
    )
    source: str = Field(description=_RecurringItemsDescriptions.source)
    notes: Optional[str] = Field(None, description=_RecurringItemsDescriptions.notes)
    amount: float = Field(description=_RecurringItemsDescriptions.amount)
    category_id: Optional[int] = Field(
        None, description=_RecurringItemsDescriptions.category_id
    )
    category_group_id: Optional[int] = Field(
        None, description=_RecurringItemsDescriptions.category_group_id
    )
    is_income: bool = Field(description=_RecurringItemsDescriptions.is_income)
    exclude_from_totals: bool = Field(
        description=_RecurringItemsDescriptions.exclude_from_totals
    )
    granularity: str = Field(description=_RecurringItemsDescriptions.granularity)
    cadence: Optional[str] = Field(None)
    quantity: Optional[int] = Field(
        None, description=_RecurringItemsDescriptions.quantity
    )
    occurrences: Dict[datetime.date, List[SummarizedTransactionObject]] = Field(
        description=_RecurringItemsDescriptions.occurrences
    )
    transactions_within_range: Optional[List[SummarizedTransactionObject]] = Field(
        None, description=_RecurringItemsDescriptions.transactions_within_range
    )
    missing_dates_within_range: Optional[List[Any]] = Field(
        None, description=_RecurringItemsDescriptions.missing_dates_within_range
    )
    date: Optional[datetime.date] = Field(
        None, description=_RecurringItemsDescriptions.date
    )
    to_base: float = Field(description=_RecurringItemsDescriptions.to_base)


class RecurringItemsParamsGet(LunchableModel):
    """
    https://lunchmoney.dev/#get-recurring-items
    """

    start_date: datetime.date
    debit_as_negative: Optional[bool] = None


class RecurringItemsClient(LunchMoneyAPIClient):
    """
    Lunch Money Recurring Items Interactions
    """

    def get_recurring_items(
        self,
        start_date: Optional[datetime.date] = None,
        debit_as_negative: Optional[bool] = None,
    ) -> List[RecurringItemsObject]:
        """
        Get Recurring Items

        Use this to retrieve a list of recurring items to expect for a specified month.

        A different set of recurring items is expected every month. These can be once a year,
        twice a year, every four months, etc.

        If a recurring item is listed as “twice a month,” then the recurring item object returned
        will have an occurrences attribute populated by the different billing dates the system believes
        recurring transactions should occur, including the two dates in the current month, the last
        transaction date prior to the month, and the next transaction date after the month.

        If the recurring item is listed as “once a week,” then the recurring item object returned will
        have an occurrences object populated with as many times as there are weeks for the specified
        month, along with the last transaction from the previous month and the next transaction for
        the next month.

        In the same vein, if a recurring item that began last month is set to “Every 3 months”,
        then that recurring item object that occurred will not include any dates for this month.

        Parameters
        ----------
        start_date : Optional[datetime.date]
            Date to search. Whatever your start date, the system will automatically
            return recurring items expected for that month. For instance, if you
            input 2020-01-25, the system will return recurring items which are to
            be expected between 2020-01-01 to 2020-01-31. By default will return
            the first day of the current month
        debit_as_negative: bool
            Pass in true if you'd like items to be returned as negative amounts
            and credits as positive amounts. Defaults to false.

        Returns
        -------
        List[RecurringItemsObject]
        """
        if start_date is None:
            start_date = datetime.datetime.now().date().replace(day=1)
        params = RecurringItemsParamsGet(
            start_date=start_date, debit_as_negative=debit_as_negative
        ).model_dump(exclude_none=True)
        response_data = self.make_request(
            method=self.Methods.GET,
            url_path=[APIConfig.LUNCH_MONEY_RECURRING_ITEMS],
            params=params,
        )
        recurring_expenses_objects = [
            RecurringItemsObject.model_validate(item) for item in response_data
        ]
        logger.debug(
            "%s RecurringExpensesObjects retrieved", len(recurring_expenses_objects)
        )
        return recurring_expenses_objects



"""
Test cases for the __main__ module.
"""

import pytest
from click.testing import CliRunner


@pytest.fixture
def runner() -> CliRunner:
    """
    Fixture for invoking command-line interfaces.
    """
    return CliRunner()


@pytest.mark.filterwarnings("ignore:datetime.datetime.utcfromtimestamp")
def test_main_succeeds(runner: CliRunner) -> None:
    """
    It exits with a status code of zero.
    """
    from lunchable._cli import cli

    result = runner.invoke(cli)
    assert result.exit_code == 0


def test_registered_plugins(runner: CliRunner) -> None:
    """
    Assert that all registered plugins are available.
    """
    from lunchable._cli import cli

    builtin_plugins = ["primelunch", "splitlunch", "pushlunch"]

    for plugin in builtin_plugins:
        result = runner.invoke(cli, ["plugins", plugin, "--help"])
        assert result.exit_code == 0



"""
Run Tests on the Recurring Items Endpoint
"""

import datetime
import logging

from lunchable import LunchMoney
from lunchable.models.recurring_items import RecurringItemsObject
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_get_recurring_items(
    lunch_money_obj: LunchMoney, obscure_start_date: datetime.datetime
):
    """
    Get Recurring Items, and ensure they are returned as RecurringItemsObject
    """
    recurring_expenses = lunch_money_obj.get_recurring_items(
        start_date=obscure_start_date
    )
    assert len(recurring_expenses) >= 1
    for recurring_expense in recurring_expenses:
        assert isinstance(recurring_expense, RecurringItemsObject)
    logger.info("%s Recurring Expenses returned", len(recurring_expenses))



"""
Lunch Money - Recurring Expenses

https://lunchmoney.dev/#recurring-expenses
"""

import datetime
import logging
import warnings
from typing import List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient
from lunchable.models._descriptions import _RecurringExpensesDescriptions

logger = logging.getLogger(__name__)


class RecurringExpensesObject(LunchableModel):
    """
    Recurring Expenses Object

    https://lunchmoney.dev/#recurring-expenses-object
    """

    id: int = Field(description=_RecurringExpensesDescriptions.id)
    start_date: Optional[datetime.date] = Field(
        None, description=_RecurringExpensesDescriptions.start_date
    )
    end_date: Optional[datetime.date] = Field(
        None, description=_RecurringExpensesDescriptions.end_date
    )
    cadence: str = Field(description=_RecurringExpensesDescriptions.cadence)
    payee: str = Field(description="Payee of the recurring expense")
    amount: float = Field(description=_RecurringExpensesDescriptions.amount)
    currency: str = Field(
        max_length=3, description=_RecurringExpensesDescriptions.currency
    )
    description: Optional[str] = Field(
        None, description=_RecurringExpensesDescriptions.description
    )
    billing_date: datetime.date = Field(
        description=_RecurringExpensesDescriptions.billing_date
    )
    type: str = Field(description=_RecurringExpensesDescriptions.type)
    original_name: Optional[str] = Field(
        None, description=_RecurringExpensesDescriptions.original_name
    )
    source: str = Field(description=_RecurringExpensesDescriptions.source)
    plaid_account_id: Optional[int] = Field(
        None, description=_RecurringExpensesDescriptions.plaid_account_id
    )
    asset_id: Optional[int] = Field(
        None, description=_RecurringExpensesDescriptions.asset_id
    )
    category_id: Optional[int] = Field(
        None, description=_RecurringExpensesDescriptions.category_id
    )


class RecurringExpenseParamsGet(LunchableModel):
    """
    https://lunchmoney.dev/#get-recurring-expenses
    """

    start_date: datetime.date
    debit_as_negative: Optional[bool] = None


class RecurringExpensesClient(LunchMoneyAPIClient):
    """
    Lunch Money Recurring Expenses Interactions
    """

    def get_recurring_expenses(
        self,
        start_date: Optional[datetime.date] = None,
        debit_as_negative: Optional[bool] = None,
    ) -> List[RecurringExpensesObject]:
        """
        Get Recurring Expenses

        **DEPRECATED** - Use [LunchMoney.get_recurring_items()][lunchable.LunchMoney.get_recurring_items]
        instead.

        Retrieve a list of recurring expenses to expect for a specified period.

        Every month, a different set of recurring expenses is expected. This is because recurring
        expenses can be once a year, twice a year, every 4 months, etc.

        If a recurring expense is listed as “twice a month”, then that recurring expense will be
        returned twice, each with a different billing date based on when the system believes that
        recurring expense transaction is to be expected. If the recurring expense is listed as
        “once a week”, then that recurring expense will be returned in this list as many times as
        there are weeks for the specified month.

        In the same vein, if a recurring expense that began last month is set to “Every 3
        months”, then that recurring expense will not show up in the results for this month.

        Parameters
        ----------
        start_date : Optional[datetime.date]
            Date to search. By default will return the first day of the current month
        debit_as_negative: bool
            Pass in true if you'd like expenses to be returned as negative amounts and credits as
            positive amounts.

        Returns
        -------
        List[RecurringExpensesObject]
        """
        warnings.warn(
            message=(
                "`LunchMoney.get_recurring_expenses` is deprecated, "
                "use `LunchMoney.get_recurring_items` instead"
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )
        if start_date is None:
            start_date = datetime.datetime.now().date().replace(day=1)
        params = RecurringExpenseParamsGet(
            start_date=start_date, debit_as_negative=debit_as_negative
        ).model_dump(exclude_none=True)
        response_data = self.make_request(
            method=self.Methods.GET,
            url_path=[APIConfig.LUNCH_MONEY_RECURRING_EXPENSES],
            params=params,
        )
        recurring_expenses = response_data.get(APIConfig.LUNCH_MONEY_RECURRING_EXPENSES)
        recurring_expenses_objects = [
            RecurringExpensesObject.model_validate(item) for item in recurring_expenses
        ]
        logger.debug(
            "%s RecurringExpensesObjects retrieved", len(recurring_expenses_objects)
        )
        return recurring_expenses_objects



"""
Run Tests on the Recurring Expenses Endpoint
"""

import datetime
import logging

import pytest

from lunchable import LunchMoney
from lunchable.models.recurring_expenses import RecurringExpensesObject
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_get_recurring_expenses(
    lunch_money_obj: LunchMoney, obscure_start_date: datetime.datetime
):
    """
    Get Recurring Expense and Assert it's a Recurring Expense
    """
    with pytest.warns(DeprecationWarning):
        recurring_expenses = lunch_money_obj.get_recurring_expenses(
            start_date=obscure_start_date
        )
    assert len(recurring_expenses) >= 1
    for recurring_expense in recurring_expenses:
        assert isinstance(recurring_expense, RecurringExpensesObject)
    logger.info("%s Recurring Expenses returned", len(recurring_expenses))



"""
Lunchmoney SDK Core
"""

from __future__ import annotations

from functools import cached_property
from typing import (
    Any,
    AsyncIterable,
    Iterable,
    Mapping,
    Optional,
    Union,
)

import httpx
import pydantic_core
from httpx import Client

from lunchable._config import APIConfig
from lunchable.exceptions import LunchMoneyHTTPError


class LunchMoneyClient(Client):
    """
    API HTTP Client
    """

    def __init__(self, access_token: str | None = None) -> None:
        timeout = httpx.Timeout(connect=5, read=30, write=20, pool=5)
        super().__init__(timeout=timeout)
        api_headers = APIConfig.get_header(access_token=access_token)
        self.headers.update(api_headers)


class LunchMoneyAsyncClient(httpx.AsyncClient):
    """
    API Async HTTP Client
    """

    def __init__(self, access_token: str | None = None) -> None:
        timeout = httpx.Timeout(connect=5, read=30, write=20, pool=5)
        super().__init__(timeout=timeout)
        api_headers = APIConfig.get_header(access_token=access_token)
        self.headers.update(api_headers)


class LunchMoneyAPIClient:
    """
    Core API Client Class
    """

    class Methods:
        """
        HTTP Request Method Enumerations: GET, OPTIONS, HEAD, POST, PUT, PATCH, or DELETE
        """

        # This Helper Namespace Organizes and Tracks HTTP Requests by Method
        GET = "GET"
        OPTIONS = "OPTIONS"
        HEAD = "HEAD"
        POST = "POST"
        PUT = "PUT"
        PATCH = "PATCH"
        DELETE = "DELETE"

    def __init__(self, access_token: str | None = None) -> None:
        """
        Initialize a Lunch Money object with an Access Token.

        Tries to inherit from the Environment if one isn't provided

        Parameters
        ----------
        access_token: Optional[str]
            Lunchmoney Developer API Access Token
        """
        self.access_token = APIConfig.get_access_token(access_token=access_token)

    def __repr__(self) -> str:
        """
        String Representation

        Returns
        -------
        str
        """
        return "<LunchMoney: httpx.Client>"

    @cached_property
    def session(self) -> httpx.Client:
        """
        Lunch Money HTTPX Client

        Returns
        -------
        httpx.Client
        """
        return LunchMoneyClient(access_token=self.access_token)

    @cached_property
    def async_session(self) -> httpx.AsyncClient:
        """
        Lunch Money HTTPX Async Client

        Returns
        -------
        httpx.AsyncClient
        """
        return LunchMoneyAsyncClient(access_token=self.access_token)

    def request(
        self,
        method: str,
        url: Union[httpx.URL, str],
        *,
        content: Optional[
            Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]]
        ] = None,
        data: Optional[Mapping[str, Any]] = None,
        json: Optional[Any] = None,
        params: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ) -> httpx.Response:
        """
        Make an HTTP request

        This is a simple method :class:`.LunchMoney` exposes to make HTTP requests. It
        has the benefit of using an existing `httpx.Client` as well as as out of the box
        auth headers that are used to connect to the Lunch Money Developer API.

        Parameters
        ----------
        method: str
            requests method: GET, OPTIONS, HEAD, POST, PUT,
            PATCH, or DELETE
        url: Union[httpx.URL, str]
            URL for the new Request object.
        content: Optional[Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]]]
            Content to send in the body of the Request.
        data: Optional[Mapping[str, Any]]
            Dictionary, list of tuples, bytes, or file-like object to send
            in the body of the Request.
        json: Optional[Any]
            A JSON serializable Python object to send in the body of the Request.
        params: Optional[Mapping[str, Any]]
            Dictionary, list of tuples or bytes to send in the query
            string for the Request.
        **kwargs: Any
            Additional arguments to send to the request method.

        Returns
        -------
        httpx.Response

        Examples
        --------
        A recent use of this method was to delete a Tag (which isn't available via the
        Developer API yet)

        ```python
        import lunchable

        lunch = lunchable.LunchMoney()

        # Get All the Tags
        all_tags = lunch.get_tags()
        # Get All The Null Tags (a list of 1 or zero)
        null_tags = [tag for tag in all_tags if tag.name in [None, ""]]

        # Create a Cookie dictionary from a browser session
        cookies = {"cookie_keys": "cookie_values"}
        del lunch.session.headers["authorization"]

        for null_tag in null_tags:
            # use the httpx.client embedded in the class to make a request with cookies
            response = lunch.request(
                method=lunch.Methods.DELETE,
                url=f"https://api.lunchmoney.app/tags/{null_tag.id}",
                cookies=cookies
            )
            # raise an error for 4XX responses
            response.raise_for_status()
        ```
        """
        response = self.session.request(
            method=method,
            url=url,
            content=content,
            data=data,
            json=json,
            params=params,
            **kwargs,
        )
        return response

    async def arequest(
        self,
        method: str,
        url: Union[httpx.URL, str],
        *,
        content: Optional[
            Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]]
        ] = None,
        data: Optional[Mapping[str, Any]] = None,
        json: Optional[Any] = None,
        params: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ) -> httpx.Response:
        """
        Make an async HTTP request

        This is a simple method :class:`.LunchMoney` exposes to make HTTP requests. It
        has the benefit of using an existing `httpx.Client` as well as as out of the box
        auth headers that are used to connect to the Lunch Money Developer API.

        Parameters
        ----------
        method: str
            requests method: GET, OPTIONS, HEAD, POST, PUT,
            PATCH, or DELETE
        url: Union[httpx.URL, str]
            URL for the new Request object.
        content: Optional[Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]]]
            Content to send in the body of the Request.
        data: Optional[Mapping[str, Any]]
            Dictionary, list of tuples, bytes, or file-like object to send
            in the body of the Request.
        json: Optional[Any]
            A JSON serializable Python object to send in the body of the Request.
        params: Optional[Mapping[str, Any]]
            Dictionary, list of tuples or bytes to send in the query
            string for the Request.
        **kwargs: Any
            Additional arguments to send to the request method.

        Returns
        -------
        httpx.Response
        """
        response = self.async_session.request(
            method=method,
            url=url,
            content=content,
            data=data,
            json=json,
            params=params,
            **kwargs,
        )
        return await response

    @classmethod
    def process_response(cls, response: httpx.Response) -> Any:
        """
        Process a Lunch Money response and raise any errors

        This includes 200 responses that are actually errors

        Parameters
        ----------
        response: httpx.Response
            An HTTPX Response Object
        """
        try:
            response.raise_for_status()
        except httpx.HTTPError as he:
            raise LunchMoneyHTTPError(response.text) from he
        if response.content:
            returned_data = response.json()
        else:
            returned_data = None
        if isinstance(returned_data, dict) and any(
            ["error" in returned_data.keys(), "errors" in returned_data.keys()]
        ):
            try:
                errors = returned_data["error"]
            except KeyError:
                errors = returned_data["errors"]
            raise LunchMoneyHTTPError(errors)
        return returned_data

    def make_request(
        self,
        method: str,
        url_path: Union[list[Union[str, int]], str, int],
        params: Optional[Mapping[str, Any]] = None,
        payload: Optional[Any] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Make an HTTP request and `process` its response

        This method is a wrapper around :meth:`.LunchMoney.request` that
        also processes the response and checks for any errors.

        Parameters
        ----------
        method: str
            requests method: GET, OPTIONS, HEAD, POST, PUT,
            PATCH, or DELETE
        url_path: Union[List[Union[str, int]], str, int]
            URL components to make into a URL
        payload: Optional[Mapping[str, Any]]
            Data to send in the body of the Request.
        params: Optional[Mapping[str, Any]]
            Dictionary, list of tuples or bytes to send in the query
            string for the Request.
        **kwargs: Any
            Additional arguments to send to the request method.

        Returns
        -------
        Any
        """
        url = APIConfig.make_url(url_path=url_path)
        json_safe_payload = pydantic_core.to_json(payload) if payload else None
        json_safe_params = pydantic_core.to_jsonable_python(params)
        response = self.request(
            method=method,
            url=url,
            params=json_safe_params,
            content=json_safe_payload,
            **kwargs,
        )
        data = self.process_response(response=response)
        return data

    async def amake_request(
        self,
        method: str,
        url_path: Union[list[Union[str, int]], str, int],
        params: Optional[Mapping[str, Any]] = None,
        payload: Optional[Any] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Make an async HTTP request and `process` its response

        This method is a wrapper around :meth:`.LunchMoney.arequest` that
        also processes the response and checks for any errors.

        Parameters
        ----------
        method: str
            requests method: GET, OPTIONS, HEAD, POST, PUT,
            PATCH, or DELETE
        url_path: Union[List[Union[str, int]], str, int]
            URL components to make into a URL
        payload: Optional[Mapping[str, Any]]
            Data to send in the body of the Request.
        params: Optional[Mapping[str, Any]]
            Dictionary, list of tuples or bytes to send in the query
            string for the Request.
        **kwargs: Any
            Additional arguments to send to the request method.

        Returns
        -------
        Any
        """
        url = APIConfig.make_url(url_path=url_path)
        json_safe_payload = pydantic_core.to_jsonable_python(payload)
        json_safe_params = pydantic_core.to_jsonable_python(params)
        response = await self.arequest(
            method=method,
            url=url,
            params=json_safe_params,
            data=json_safe_payload,
            **kwargs,
        )
        data = self.process_response(response=response)
        return data



"""
Run Tests on the Tags Endpoint
"""

import logging

from lunchable import LunchMoney
from lunchable.models.tags import TagsObject
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_get_tags(lunch_money_obj: LunchMoney):
    """
    Get Plaid Account and Assert it's a Plaid Account
    """
    tags = lunch_money_obj.get_tags()
    assert len(tags) >= 1
    for tag in tags:
        assert isinstance(tag, TagsObject)
    logger.info("%s Plaid Accounts returned", len(tags))
